"""
pansat.download.providers.laads_daac
====================================

This module provides a data provider for files from the The Level-1 and
Atmosphere Archive & Distribution System (LAADS) Distributed Active Archive
Center (DAAC).

Reference
---------
"""
from pansat.download import accounts
from pansat.download.providers.discrete_provider import DiscreteProvider
import requests
import re

LAADS_PRODUCTS = [
    "MODIS_Terra_MOD021KM",
    "MODIS_Terra_MOD03",
    "MODIS_Terra_MOD35_l2",
    "MODIS_Aqua_MYD021KM",
    "MODIS_Aqua_MYD03",
    "MODIS_Aqua_MYD35_l2",
]


class LAADSDAACProvider(DiscreteProvider):
    """
    Dataprovider class for for products available from the
    gpm1.gesdisc.eosdis.nasa.gov domain.
    """

    base_url = "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/61/"
    file_pattern = re.compile("[\w\.]*.hdf")

    def __init__(self, product):
        """
        Create new GesDisc provider.

        Args:
            product: The product to download.
        """
        super().__init__(product)

    @classmethod
    def get_available_products(cls):
        """
        Return the names of products available from this data provider.

        Return:
            A list of strings containing the names of the products that can
            be downloaded from this data provider.
        """
        return LAADS_PRODUCTS

    @property
    def _request_string(self):
        """The URL containing the data files for the given product."""
        base_url = LAADSDAACProvider.base_url
        base_url += f"{self.product.product_name.upper()}"
        return base_url + "/{year}/{day}/{filename}"

    def get_files_by_day(self, year, day):
        """
        Return list of available files for a given day of a year.

        Args:
            year(``int``): The year for which to look up the files.
            day(``int``): The Julian day for which to look up the files.

        Return:
            A list of strings containing the filename that are available
            for the given day.
        """
        day = str(day)
        day = "0" * (3 - len(day)) + day
        request_string = self._request_string.format(year=year, day=day, filename="")
        response = requests.get(request_string)
        files = list(set(LAADSDAACProvider.file_pattern.findall(response.text)))
        return [f for f in files]

    def download_file(self, filename, destination):
        """
        Download file from data provider.

        Args:
            filename(``str``): The name of the file to download.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """
        t = self.product.filename_to_date(filename)
        year = t.year
        day = t.strftime("%j")
        day = "0" * (3 - len(day)) + day
        request_string = self._request_string.format(
            year=year, day=day, filename=filename
        )

        auth = accounts.get_identity("GES DISC")
        response = requests.get(request_string, auth=auth)
        url = response.url
        response = requests.get(url, auth=auth)

        with open(destination, "wb") as f:
            for chunk in response:
                f.write(chunk)



"""
pansat.download
===============

The :py:mod:`pansat.download` module contains all code responsible for
downloading data. It is itself structured into three parts with
different responsibilities:

   1. The :py:mod:`pansat.download.providers` module, which contains all data providers.
      Data provider class provide interfaces for different source from which data
      can be downloaded.
   2. The :py:mod:`pansat.download.accounts` module, which manages the login information
      for different data providers.
   3. The :py:mod:`pansat.download.commandline` module, which provides a command line
      interface for the download functionality.
"""



"""
pansat.products.product
=======================

This module provides the Product abstract base class which defines the general
interface for objects representing data products.
"""
from abc import ABC, abstractproperty, abstractmethod


class Product(ABC):
    """
    The abstract interface for satellite and reanalysis data products.
    """

    @abstractproperty
    def default_destination(self):
        """Default folder structure used to store results."""

    @abstractmethod
    def filename_to_date(self, filename):
        """
        Extract data from filename.

        Args:
            filename(``str``): The filename.

        Returns:
            ``datetime`` object corresponding to the date encoded in the
            filename.
        """

    @abstractmethod
    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start date
                 of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date of the
                 of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to store
                 the output data.
        """

    @abstractmethod
    def __str__(self):
        """Should return a string representation of the product name."""



"""
pansat.formats.hdf4
===================

This module provides a wrapper for the ``pyhdf`` package to simplify reading
of HDF4 files. The main interface to read a HDF4 file is implemented by the
``HDF4File`` class. The ``Dataset`` and ``VData`` classes represent the types
of variables that a HDF4 file contains.

Example
-------

With the ``HDF4File`` class reading HDF4 files and accessing their variables is
as simple as shown below:

.. code-block::

    file = HDF4File("file.hdf")
    print(file.variables)      # Print all variables in file.
    data = file.variable_1[:]  # Read data from variable named `variable_1`

"""
import weakref
import numpy as np
import logging

LOGGER = logging.getLogger(__name__)

try:
    from pyhdf.HDF import HDF
    from pyhdf.SD import SD
    from pyhdf.VS import VS
except ImportError as error:
    LOGGER.error("The pyhdf package is required to read HDF4 file. Please install it.")
    raise error


class VData:
    """
    Class representing VData objects, i.e. numeric data that is stored in table
    format in an HDF file and accessed through the VS interface.

    Attributes:
        file(``weakref``): Weak reference to file object required for data
            access.
        name(``str``): The name of the attribute
        cls(``str``): The attribute class
        reference(``int``): Reference number identifying the vdata object.
        n_records(``int``): The number of records, i.e. rows, of the data
            table.
        n_fields(``int``): The number of fields, i.e. columns, of the
            data table.
        n_attributes(``int``): The number of attributes.
        size(``int``): Size of a single record (row) in bytes.
        tag(``int``): The vdata tag number.
        interlace(``int``): The vdata interlace mode.
    """

    def __init__(
        self,
        file,
        name,
        cls,
        reference,
        n_records,
        n_fields,
        n_attributes,
        size,
        tag,
        interlace,
    ):
        self.file = file
        self.name = name
        self.cls = cls
        self.reference = reference
        self.n_records = n_records
        self.n_fields = n_fields
        self.n_attributes = n_attributes
        self.size = size
        self.tag = tag
        self.interlace = interlace

    def __repr__(self):
        return f"VData({self.name}, [{self.n_records}, {self.n_fields}])"

    def __str__(self):
        return f"HDF4 VData object: {self.name}, records={self.n_records}"

    def __getitem__(self, *args):
        """
        Selects datasets from file and forwards call to the returned vdata
        object.
        """
        data = self.file().vdata_table.attach(self.name).__getitem__(*args)
        return np.array(data)


class Dataset:
    """
    Class representing HDF4 Datasets, i.e. numeric data that is stored as
    multi-dimensional array and accessed through the SD interface.

    Attributes:
        file(``weakref``): Weak reference to file object required for data
            access.
        name(``str``): The name of the dataset.
        dimensions(``tuple``): Tuple containing the variable names of the
            dimensions holding the dimensions of the dataset.
        shape(``tuple``): Tuple containing the shape of the dataset.
        hdf_type(``int``): Integer representing the HDF-internal type of the
            dataset
        index(``int``): Integer representing the HDF-internal index of the
            dataset.
    """

    def __init__(self, file, name, dimensions, shape, hdf_type, index):
        self.file = file
        self.name = name
        self.dimensions = dimensions
        self.shape = shape
        self.hdf_type = hdf_type
        self.index = index

    def __repr__(self):
        return f"Dataset({self.name}, {self.dimensions}, {self.shape})"

    def __str__(self):
        return f"HDF4 Dataset object: {self.name}"

    def __getitem__(self, *args):
        """
        Selects datasets from file and forwards call to the returned dataset
        object.
        """
        file = self.file()
        return file.scientific_dataset.select(self.name).__getitem__(*args)


class HDF4File:
    """
    Simplified interface for reading HDF4 files. It combines the SD and VS
    low-level interfaces.
    """

    # Attributes:
    #     variables(``list``): List of strings of variable names contained in
    #         this file.

    def __init__(self, path):
        self.path = path
        self.file_handle = HDF(str(path))

        self.scientific_dataset = SD(str(path))
        datasets = self.scientific_dataset.datasets()
        dataset_dict = {
            key: Dataset(weakref.ref(self), key, *info)
            for key, info in datasets.items()
        }
        self.datasets = dataset_dict

        self.vdata_table = VS(self.file_handle)
        vdata_dict = {
            info[0]: VData(weakref.ref(self), *info)
            for info in self.vdata_table.vdatainfo()
        }
        self.vdata = vdata_dict

    def __del__(self):
        if self.file_handle:
            self.file_handle.close()
            self.file_handle = None

    @property
    def variables(self):
        """
        Names of the variables available in this file.
        """
        return list(self.datasets.keys()) + list(self.vdata.keys())

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError as error:
            datasets = object.__getattribute__(self, "datasets")
            if name in datasets:
                return datasets[name]
            vdata = object.__getattribute__(self, "vdata")
            if name in vdata:
                return vdata[name]
            raise error

    def __repr__(self):
        return f"HDF4File({self.path})"



"""
pansat.products.satellite.persiann
==================================

This module provides and interface to download and open data from the
PERSIANN suite or precipitation retrieval algorithms.
"""
from datetime import datetime, timedelta
import gzip
from pathlib import Path

import numpy as np
import xarray as xr

import pansat.download.providers as providers
from pansat.products.product import Product



class PersiannProduct(Product):
    """
    Base class for PERSIANN precipitation products.
    """
    def __init__(self, resolution=1):
        self.resolution = resolution

    def filename_to_date(self, filename):
        """
        Determine the data from a given filename.

        Args:
            filename: The name of the file of which to determine the data.

        Return:
            A 'datetime.datetime' object representing the date corresponding
            to the given file.
        """
        filename = str(filename)
        year = 2000 + int(filename[-14:-12])
        day_of_year = int(filename[-12:-9])
        hour = int(filename[-9:-7])
        date = (datetime(year=year, month=1, day=1, hour=hour) +
                timedelta(days=day_of_year - 1))
        return date

    def open(self, filename):
        """
        Open file as 'xarray.Dataset'.

        Args:
            filename: Path to the file to open.

        Return:
            An 'xarray.Dataset' containing the data from the given
            file.
        """
        bytes = gzip.open(filename).read()
        shape = (3000, 9000)

        data = np.frombuffer(bytes, ">i2").reshape(shape)
        lons = np.linspace(0.02, 359.98, 9000)
        lats = np.linspace(59.98, -59.98, 3000)

        date = self.filename_to_date(filename)

        data = data / 100
        data[data < 0] = np.nan

        return xr.Dataset({
            "time": (("time",), [date]),
            "latitude": (("latitude",), lats),
            "longitude": (("longitude",), lons),
            "precipitation": (("time", "latitude", "longitude"),
                              data[np.newaxis])
        })

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {str(self)}."
            )
        return available_providers[0]

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)


class CCS(PersiannProduct):
    """
    The PERSIANN cloud classification system (CCS) precipitation retrieval.
    """
    RESOLUTIONS = {
        1: "hrly",
        3: "3hrly",
        6: "6hrly",
        24: "daily"
    }

    def __init__(self, resolution=1):
        super().__init__(resolution=resolution)

    @property
    def default_destination(self):
        """
        Files are stored in a 'PERSIANN-CCS' sub-folder.
        """
        return Path("PERSIANN-CCS")

    def __str__(self):
        return "PERSIANN-CCS"

    def get_path(self, year):
        """
        Use by the provider to determine the folder in which the
        files corresponding this product are found.
        """
        path = f"PERSIANN-CCS/{CCS.RESOLUTIONS[self.resolution]}"
        if self.resolution == 1:
            path = path + f"/{year}"
        return path


class PDIRNow(PersiannProduct):
    """
    The PERSIANN cloud classification system (CCS) precipitation retrieval.
    """
    RESOLUTIONS = {
        1: "1hourly",
        3: "3hourly",
        6: "6hourly"
    }
    def __init__(self, resolution=1):
        super().__init__(resolution=resolution)

    @property
    def default_destination(self):
        """
        Files are stored in a 'PERSIANN-CCS' sub-folder.
        """
        return Path("PDIRNow")

    def filename_to_date(self, filename):
        """
        Determine the data from a given filename.

        Args:
            filename: The name of the file of which to determine the data.

        Return:
            A 'datetime.datetime' object representing the date corresponding
            to the given file.
        """
        filename = str(filename)
        year = 2000 + int(filename[-15:-13])
        month = int(filename[-13:-11])
        day = int(filename[-11:-9])
        hour = int(filename[-9:-7])
        date = datetime(year=year, month=month, day=day, hour=hour)
        return date

    def __str__(self):
        return "PDIRNow"

    def get_path(self, year):
        """
        Use by the provider to determine the folder in which the
        files corresponding this product are found.
        """
        path = f"PDIRNow/PDIRNow{PDIRNow.RESOLUTIONS[self.resolution]}"
        if self.resolution == 1:
            path = path + f"/{year}"
        return path



"""
pansat.download.providers.noaa
==============================

This module provides the NoaaProvider class to download data stored at the NOAA data server.

"""


import ftplib
from ftplib import FTP
from pathlib import Path
import numpy as np
from pansat.download.accounts import get_identity
from pansat.download.providers.data_provider import DataProvider


NOAA_PRODUCTS = [
    "ncep.reanalysis-surface",
    "ncep.reanalysis-pressure",
    "ncep.reanalysis-surface_gauss",
    "ncep.reanalysis-spectral",
    "ncep.reanalysis-tropopause",
    "igra-soundings",
]


class NOAAProvider(DataProvider):
    """
    Abstract base class for gridded products available from NOAA Physical Science Laboratory.
    """

    base_url = "ftp2.psl.noaa.gov"

    def __init__(self, product):
        """
        Create a new product instance.

        Args:

            product: Product class object with specific product for NOAA
        """
        super().__init__()
        self.product = product
        self.product_path = "/Datasets/" + ("/").join(self.product.name.split("-"))

        self.cache = {}

        if not product.name in NOAA_PRODUCTS:
            available_products = NOAA_PRODUCTS
            raise ValueError(
                f"{product.name} not available from the NOAA data"
                " provider. Currently available products are: "
                f" {available_products}."
            )

    @classmethod
    def get_available_products(cls):
        """
        Return the names of products available from this data provider.

        Return:
            A list of strings containing the names of the products that can
            be downloaded from this data provider.
        """
        return NOAA_PRODUCTS

    def _ftp_listing_to_list(self, path, item_type=int, base_url=None):
        """
        Retrieve directory content from ftp listing as list.

        Args:

            path(str): The path from which to retrieve the ftp listing.
            t(type): Type constructor to apply to the elements of the
                listing. To retrieve a list of strings use t = str.
            base_url(``str``): FTP URL without subdirectories

        Return:

            A list containing the content of the ftp directory.

        """

        if base_url is None:
            base_url = NOAAProvider.base_url

        if not path in self.cache:
            with FTP(base_url) as ftp:
                user, password = get_identity("NOAAProvider")
                ftp.login(user=user, passwd=password)
                try:
                    ftp.cwd(path)
                except:
                    raise Exception(
                        "Can't find product folder "
                        + path
                        + "on the NOAA server. Are you sure this is the right path?"
                    )
                listing = ftp.nlst()
            listing = [item_type(l) for l in listing]
            self.cache[path] = listing
        return self.cache[path]

    @classmethod
    def get_file_names(cls, var, start, end):
        """
        Return all files from given year.

        Args:
            var(``str``): Variable to extract
            start(``int``): start year for desired time range
            end(``int``): end year for desired timerange
        Return:
            List of the filenames of this product for given variable and time range by year."""

        files = []
        for year in np.arange(start, end + 1):
            year_str = str(year)
            filename = var + "." + year_str + ".nc"
            files.append(filename)
        return files

    def download(
        self, start, end, destination, base_url=None, product_path=None, files=None
    ):
        """
        This method downloads data for a given time range from the respective
        data provider.

        Args:
            start(``int``): start year
            end(``int``): end year
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
            base_url(``str``): base url
            files(`list``): list of files if files are not sorted by year
        """

        if base_url is None:
            base_url = self.base_url
        if product_path is None:
            product_path = self.product_path

        # get file list
        if files is None:
            files = self.get_file_names(self.product.variable, start, end)

        ftp = ftplib.FTP(base_url)
        user, password = get_identity("NOAAProvider")
        ftp.login(user=user, passwd=password)
        ftp.cwd(product_path)

        output_files = []
        for filename in files:
            output = Path(str(destination)) / str(filename)
            output_files.append(str(output))
            with open(str(output), "wb") as ftpfile:
                ftp.retrbinary("RETR " + filename, ftpfile.write)

        ftp.quit()

        return output_files



"""
================
pansat.exception
================

This module defines a hierarchy of exception classes which should be used
by pansat code.
"""


class PansatException(Exception):
    """
    Base class for all exceptions thrown from within pansat.
    """


class CommunicationError(PansatException):
    """
    This error is thrown when the communication with any of the remote
    servers fails.
    """


class NoAvailableProvider(PansatException):
    """
    Exception indicating that no suitable provider could be found for
    a product.
    """


class MissingDependency(PansatException):
    """
    Thrown when an optional dependency is found to be missing.
    """



"""
pansat.download.commandline
===============================
The ``commandline`` submodule allows to download data products from commandline using the argparse module.

The following flags can be used:
   Download flags

   .. code-block::

      --type
      --pm
      --product/-prod
      --starttime/-t0
      --endtime/-t1
      --variable/-var
      --domain/-d
      --grid
   
   Other flags

   .. code-block::

      --list
      --add
      --listIDs


"""
import argparse
import datetime
import pkgutil
import importlib
import sys
import pansat.products


def download():
    """
    The function download() passes the commandline input to the respective download functions
    of the corresponding pansat.download submodules.
    """
    ################################################################
    # define helpstrings, i.e. what is shown for the --help/-h flag
    ################################################################

    helpstring_t0 = "start of time interval in ISO 8601 format (YYYY-MM-DDThh:mm:ss)"
    helpstring_t1 = "end of time interval in ISO 8601 format (YYYY-MM-DDThh:mm:ss)"
    helpstring_type = "data type (satellite/reanalysis)"
    helpstring_pm = (
        "satellite (e.g. Cloudsat)/ model (e.g ERA5)/ ground_based/ stations"
    )
    helpstring_product = "product to download"
    helpstring_variable = "variable(s) for reanalysis/station data"
    helpstring_domain = (
        "data domain of model data in latitude and longitude, if not specified"
        + " a global grid is used, if specified four arguments (lat1, lat2, lon1, lon2)"
        + " are required, with lat1<lat2, lon1<lon2"
    )
    helpstring_grid = "specifying spatial/temporal gridding of reanalysis data"
    helpstring_loc = "location of station in [lat,lon]"
    helpstring_name = "name of station"
    helpstring_add = "add account, requires arguments ``provider´´ and ``user_name´´"

    parser = argparse.ArgumentParser()

    ##############################################################################################
    # definition of flags
    ##############################################################################################
    parser.add_argument(
        "--list", action="store_true", help="list available providers/products"
    )
    parser.add_argument("--listIDs", action="store_true", help="list stored identities")
    parser.add_argument("--add", nargs=2, help=helpstring_add)

    parser.add_argument(
        "-t0",
        "--starttime",
        type=datetime.datetime.fromisoformat,
        help=helpstring_t0,
    )
    parser.add_argument(
        "-t1", "--endtime", type=datetime.datetime.fromisoformat, help=helpstring_t1
    )

    parser.add_argument(
        "--type",
        choices=["satellite", "reanalysis", "ground_based", "stations"],
        help=helpstring_type,
    )
    parser.add_argument("--pm", help=helpstring_pm)
    parser.add_argument("-prod", "--product", help=helpstring_product)
    parser.add_argument("-var", "--variable", nargs="+", help=helpstring_variable)
    parser.add_argument("--grid", nargs="+", help=helpstring_grid)
    parser.add_argument("-d", "--domain", nargs=4, type=float, help=helpstring_domain)
    parser.add_argument("-loc", "--location", nargs=2, type=float, help=helpstring_loc)
    parser.add_argument("--name", help=helpstring_name)
    parser.add_argument(
        "--recent", action="store_true", help="only download last month of IGRA data"
    )
    args = parser.parse_args()

    if args.list:
        modnames = []
        package = pansat.download
        modlen = len("pansat.download.providers.")
        for importer, modname, ispkg in pkgutil.walk_packages(
            path=package.__path__, prefix=package.__name__ + ".", onerror=lambda x: None
        ):
            if "providers." in modname and "_provider" not in modname:
                modnames.append(modname)
                print("Provider: " + str(modname[modlen:]))
                module = importlib.import_module(modname)
                products = getattr(module, str(modname[modlen:]).upper() + "_PRODUCTS")

                print("available products:")
                if isinstance(products, dict):
                    for k, v in products.items():
                        print(k, v)
                else:
                    for i in range(0, len(products)):
                        print(products[i])
        return

    if args.add:
        from pansat.download import accounts

        accounts.add_identity(args.add[0], args.add[1])
        return

    if args.listIDs:
        from pansat.download import accounts

        identities = accounts.get_identities()
        print("Identities are available for the following providers")
        for key in identities.keys():
            if key != "pansat":
                print(key)
        return

    #################################################################################
    # consistency checks to begin downloads
    #################################################################################

    if (args.starttime and not args.endtime) or (args.endtime and not args.starttime):
        parser.error("both -t0 and -t1 need to be specified at the same time")

    if not args.pm:
        parser.error("no platform/model chosen")

    if str(args.type) == "satellite" and not args.product:
        parser.error("no satellite data product chosen")

    if str(args.type) == "reanalysis" and not args.variable:
        parser.error("no reanalysis variable chosen")

    if str(args.type) == "reanalysis" and not args.grid:
        parser.error("no grid variable for reanalysis data chosen")

    if str(args.pm) == "ncep" and len(args.grid) != 1:
        parser.error("NCEP data requires exactly one argument for gridding information")

    if str(args.pm) == "ERA5" and len(args.grid) != 2:
        parser.error(
            "ERA5 data requires exactly two arguments for gridding information"
        )

    ##############################################################################################
    # loading download functions and starting download
    ##############################################################################################
    modnames = []
    package = pansat.products
    for importer, modname, ispkg in pkgutil.walk_packages(
        path=package.__path__, prefix=package.__name__ + ".", onerror=lambda x: None
    ):
        modnames.append(modname)

    modnames.append("pansat.products.ground_based.opera")
    modnames.append("pansat.products.stations.igra")

    if "pansat.products." + str(args.type) + "." + str(args.pm) in modnames:
        module = importlib.import_module(
            "pansat.products." + str(args.type) + "." + str(args.pm)
        )
        if "l" + str(args.product) in dir(module):
            productfunc = getattr(module, "l" + str(args.product))
        elif str(args.product) in dir(module):
            productfunc = getattr(module, str(args.product))
        elif "ERA5Product" in dir(module):
            era_product = getattr(module, "ERA5Product")
            if not args.domain:  # without given spatial domain (global)
                productfunc = era_product(
                    str(args.grid[0]), str(args.grid[1]), args.variable
                )
            else:
                productfunc = era_product(
                    str(args.grid[0]),
                    str(args.grid[1]),
                    args.variable,
                    args.domain,
                )
        elif "NCEPReanalysis" in dir(module):
            ncep_product = getattr(module, "NCEPReanalysis")
            productfunc = ncep_product(str(args.variable[0]), str(args.grid[0]))
        elif "IGRASoundings" in dir(module):
            igra_product = getattr(module, "IGRASoundings")
            if args.location:
                productfunc = igra_product(args.location)
            elif args.name:
                productfunc = igra_product(args.name)
            elif args.variable:
                productfunc = igra_product(variable=str(args.variable[0]))
            else:
                parser.error("can't download IGRA data with given arguments")
        else:
            parser.error("product " + str(args.product) + " not implemented")

    else:
        parser.error(
            "submodule pansat.products."
            + str(args.type)
            + "."
            + str(args.pm)
            + " is not implemented"
        )

    if args.pm == "ncep":
        args.starttime = int(args.starttime.year)
        args.endtime = int(args.endtime.year)

    if args.type != "stations":
        files = productfunc.download(args.starttime, args.endtime)
    else:
        if not args.variable:
            files = productfunc.download()
        else:
            if not args.recent:
                files = productfunc.download()
            else:
                files = productfunc.download("recent")



from datetime import datetime, timedelta
import http
import time
from pathlib import Path
from xml.etree import ElementTree

from pansat.download.providers.data_provider import DataProvider
from pansat.download.accounts import get_identity


def ensure_extension(path, ext):
    if not any([path[-len(e) :] == e for e in ext]):
        path = path + ext[0]
    return path


class GeoservicesProvider(DataProvider):
    """
    Base class for data products available from the ICARE ftp server.
    """

    base_url = "https://geoservices.meteofrance.fr/services"

    def __init__(self, product):
        """
        Create a new product instance.

        Arguments:

        product_path(str): The path of the product. This should point to
            the folder that bears the product name and contains the directory
            tree which contains the data files sorted by date.

        name_to_date(function): Funtion to convert filename to datetime object.
        """
        super().__init__()
        self.product = product
        name, password = get_identity("GeoservicesMeteofrance")
        request = (
            GeoservicesProvider.base_url
            + f"/GetAPIKey?username={name}&password={password}"
        )
        c = http.client.HTTPSConnection("geoservices.meteofrance.fr")
        c.request("GET", request)
        r = c.getresponse().read().decode()
        root = ElementTree.fromstring(r)
        self.token = root.text

    @classmethod
    def get_available_products(cls):
        return [
            "OPERA_RAINFALL_RATE",
            "OPERA_MAXIMUM_REFLECTIVITY",
            "OPERA_HOURLY_RAINFALL",
        ]

    def _get_times_in_range(self, start_time, end_time):
        seconds_since_hour = start_time.minute * 60 + start_time.second
        d_t = (15 * 60 - seconds_since_hour) % (15 * 60)
        d_t = timedelta(seconds=d_t)
        time = start_time + d_t
        while time < end_time:
            yield time
            time += timedelta(minutes=15)

    def _get_filename(self, time):
        product_name = str(self.product)[6:]
        year = time.year
        day = time.timetuple().tm_yday
        hour = time.hour
        minute = time.minute
        filename = f"OPERA_{product_name}_{year}_{day:03}_" f"{hour:02}_{minute:02}.hdf"
        return filename

    def _download_file(self, time, destination):
        """
        Download a given product file.

        Arguments:

            filename(str): The name of the file to download.

            dest(str): Where to store the file.
        """
        # if Path(destination).exists:
        #    return destination
        product_name = str(self.product)[6:]
        c = http.client.HTTPSConnection("geoservices.meteofrance.fr")
        time_str = time.strftime("%Y-%m-%dT%H:%M:%SZ")
        request = (
            GeoservicesProvider.base_url
            + f"/odyssey?product={product_name}"
            + f"&time={time_str}&token={self.token}&format=HDF5"
        )
        c.request("GET", request)
        r = c.getresponse()
        with open(destination, "wb") as f:
            f.write(r.read())
        return destination

    def download(self, start_time, end_time, destination):
        destination = Path(destination)
        files = []
        for time in self._get_times_in_range(start_time, end_time):
            filename = self._get_filename(time)
            files.append(self._download_file(time, destination / filename))
        return files

    def name_to_date(self, name):
        s = "_".join(name.split("_")[-4:])
        return datetime.strptime(s.split(".")[0], "%Y_%j_%H_%M")



"""
pansat.products.satellite.goes
==============================

This module defines the GOES product class, which is used to represent all
products from the GOES series of geostationary satellites.
"""
import re
from datetime import datetime
from pathlib import Path

import xarray

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


class GOESProduct(Product):
    """
    Base class for products from any of the currently operational
    GOES satellites (GOES 16 and 17).

    Attributes:
        series_index(``int``): Index identifying the GOES satellite
            in the GOES seris.
        level(``int``): The operational level of the product.
        name(``str``): The name of the product.
        channel(``int``): The channel index.
    """

    def __init__(self, series_index, level, name, region, channel):
        self.series_index = series_index
        self.level = level
        self.name = name
        self.region = region
        self.channel = channel
        if type(channel) == list:
            channels = "(" + "|".join([f"{c:02}" for c in channel]) + ")"

            self.filename_regexp = re.compile(
                rf"OR_ABI-L{self.level}-{self.name}{self.region}-\w\wC{channels}"
                r"_\w\w\w_s(\d*)_e(\d*)_c(\d*).nc"
            )
        else:
            self.filename_regexp = re.compile(
                rf"OR_ABI-L{self.level}-{self.name}{self.region}-\w\wC"
                rf"({self.channel:02})_\w\w\w_s(\d*)_e(\d*)_c(\d*).nc"
            )

    @property
    def variables(self):
        return []

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a GOES product.

        Returns:
            ``datetime`` object representing the timestamp of the
                filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(2)[:-1]
        date = datetime.strptime(date_string, "%Y%j%H%M%S")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for GOES product is
        ``GOES-<index>/<product_name>``>
        """
        return Path(f"GOES-{self.series_index}") / Path(str(self))

    def __str__(self):
        """The full product name."""
        return f"GOES-{self.series_index}-ABI-L{self.level}-{self.name}{self.region}"

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)

        if isinstance(self.channel, list):
            files = []
            for c in self.channel:
                prod = GOESProduct(
                    self.series_index, self.level, self.name, self.region, c
                )
                p = provider(prod)
                files += p.download(start_time, end_time, destination)
            return files

        provider = provider(self)
        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The GOES file to open.
        """
        return xarray.open_dataset(filename)


class GOES16L1BRadiances(GOESProduct):
    """
    Class representing GOES16 L1 radiance products.
    """

    def __init__(self, region, channel):
        super().__init__(16, "1b", "Rad", region, channel)


class GOES17L1BRadiances(GOESProduct):
    """
    Class representing GOES17 L1 radiance products.
    """

    def __init__(self, region, channel):
        super().__init__(17, "1b", "Rad", region, channel)


goes_16_l1b_radiances_c01_full_disk = GOES16L1BRadiances("F", 1)
goes_16_l1b_radiances_c02_full_disk = GOES16L1BRadiances("F", 2)
goes_16_l1b_radiances_c03_full_disk = GOES16L1BRadiances("F", 3)
goes_16_l1b_radiances_c04_full_disk = GOES16L1BRadiances("F", 4)
goes_16_l1b_radiances_rgb_full_disk = GOES16L1BRadiances("F", [1, 2, 3])
goes_16_l1b_radiances_all_full_disk = GOES16L1BRadiances("F", list(range(1, 17)))
goes_16_l1b_radiances_c01_conus = GOES16L1BRadiances("C", 1)
goes_16_l1b_radiances_c02_conus = GOES16L1BRadiances("C", 2)
goes_16_l1b_radiances_c03_conus = GOES16L1BRadiances("C", 3)
goes_16_l1b_radiances_c04_conus = GOES16L1BRadiances("C", 4)
goes_16_l1b_radiances_rgb_conus = GOES16L1BRadiances("C", [1, 2, 3])
goes_16_l1b_radiances_all_conus = GOES16L1BRadiances("C", list(range(16)))

goes_17_l1b_radiances_c01_full_disk = GOES17L1BRadiances("F", 1)
goes_17_l1b_radiances_c02_full_disk = GOES17L1BRadiances("F", 2)
goes_17_l1b_radiances_c03_full_disk = GOES17L1BRadiances("F", 3)
goes_17_l1b_radiances_c04_full_disk = GOES17L1BRadiances("F", 4)
goes_17_l1b_radiances_rgb_full_disk = GOES17L1BRadiances("F", [1, 2, 3])
goes_17_l1b_radiances_all_full_disk = GOES17L1BRadiances("F", list(range(1, 17)))
goes_17_l1b_radiances_c01_conus = GOES17L1BRadiances("C", 1)
goes_17_l1b_radiances_c02_conus = GOES17L1BRadiances("C", 2)
goes_17_l1b_radiances_c03_conus = GOES17L1BRadiances("C", 3)
goes_17_l1b_radiances_c04_conus = GOES17L1BRadiances("C", 4)
goes_17_l1b_radiances_rgb_conus = GOES17L1BRadiances("C", [1, 2, 3])
goes_17_l1b_radiances_all_conus = GOES17L1BRadiances("C", list(range(1, 17)))



"""
pansat.products.satellite.cloud_sat
===================================

This module defines the CloudSat product class, which represents all
supported CloudSat products.
"""
import re
import os
from datetime import datetime
from pathlib import Path

import numpy as np

from pansat.products.product_description import ProductDescription
import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


class CloudSatProduct(Product):
    """
    The CloudSat class defines a generic interface for CloudSat products.

    Attributes:
        name(``str``): The name of the product
        description(``list``): List of variable names provided by this
            product.
    """

    def __init__(self, name, description):
        self.name = name
        self._description = description
        self.filename_regexp = re.compile(
            r"([\d]*)_([\d]*)_CS_" + name + r"_GRANULE_P_R([\d]*)_E([\d]*)\.*"
        )

    @property
    def description(self):
        # Product description object describing the CloudSat product.
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a CloudSat product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        filename = os.path.basename(filename)
        filename = filename.split("_")[0]
        return datetime.strptime(filename, "%Y%j%H%M%S")

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for CloudSat product is
        ``CloudSat/<product_name>``>
        """
        return Path("CloudSat") / Path(self.name)

    def __str__(self):
        """The full product name."""
        return "CloudSat_" + self.name

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The CloudSat file to open.
        """
        from pansat.formats.hdf4 import HDF4File

        file_handle = HDF4File(filename)
        return self.description.to_xarray_dataset(file_handle, globals())


def _cloud_scenario_to_cloud_scenario_flag(cloud_scenario):
    """
    Extract cloud class from CloudSat cloud scenario data.

    Extract bits 0 from the combined integer values, which encode the
    whether the cloud type could be determined.
    """
    cloud_scenario[:].astype(np.int16)
    mask = 0x0001
    return np.bitwise_and(cloud_scenario[:], mask)


def _cloud_scenario_to_cloud_class(cloud_scenario):
    """
    Extract cloud class from CloudSat cloud scenario data.

    Extract bits 1-4 from the combined integer values, which encode the
    cloud class.
    """
    cloud_scenario[:].astype(np.int16)
    mask = 0x001E
    return np.right_shift(np.bitwise_and(cloud_scenario[:], mask), 1)


def _cloud_scenario_to_land_sea_flag(cloud_scenario):
    """
    Extract sea flag from CloudSat cloud scenario data.

    Extract bits 5-6 from the combined integer values, which encode the
    land-sea flag.
    """
    mask = 0x0060
    return np.right_shift(np.bitwise_and(cloud_scenario[:], mask), 5)


def _cloud_scenario_to_latitude_flag(cloud_scenario):
    """
    Extract latitude flag from CloudSat cloud scenario data.

    Extract bits 7-8 from the combined integer values, which encode the
    latitude flag.
    """
    mask = 0x0180
    return np.right_shift(np.bitwise_and(cloud_scenario[:], mask), 7)


def _cloud_scenario_to_algorithm_flag(cloud_scenario):
    """
    Extract algorithm flag from CloudSat cloud scenario data.

    Extract bits 9-10 from the combined integer values, which encode the
    algorithm used for classification.
    """
    mask = 0x0600
    return np.right_shift(np.bitwise_and(cloud_scenario[:], mask), 9)


def _cloud_scenario_to_quality_flag(cloud_scenario):
    """
    Extract quality flag from CloudSat cloud scenario data.

    Extract bits 11-12 from the combined integer values, which encode the
    data quality.
    """
    mask = 0x1800
    return np.right_shift(np.bitwise_and(cloud_scenario[:], mask), 11)


def _cloud_scenario_to_precipitation_flag(cloud_scenario):
    """
    Extract precipitation flag from CloudSat cloud scenario data.

    Extract bits 13-14 from the combined integer values, which encode the
    precipitation flag.
    """
    mask = 0x6000
    return np.right_shift(np.bitwise_and(cloud_scenario[:], mask), 13)


def _parse_products():
    module_path = Path(__file__).parent
    for filename in module_path.iterdir():
        if filename.match("*.ini"):
            description = ProductDescription(filename)
            python_name = filename.name.split(".")[0]
            product_name = description.name
            globals()[python_name] = CloudSatProduct(product_name, description)


_parse_products()



"""
pansat.download.providers.ges_disc
==================================

This module contains a data provider for NASA's Goddard Earth Sciences Data and
Information Services Center (`GES DISC <https://disc.gsfc.nasa.gov/>`_).

Reference
---------
"""
import datetime
import json
import os
import pathlib
import re
import tempfile

import requests

from pansat.download import accounts
from pansat.download.providers.discrete_provider import DiscreteProvider

_DATA_FOLDER = pathlib.Path(__file__).parent / "data"
with open(_DATA_FOLDER / "gpm_products.json", "r") as file:
    GPM_PRODUCTS = json.load(file)


class GesdiscProvider(DiscreteProvider):
    """
    Dataprovider class for for products available from the
    gpm1.gesdisc.eosdis.nasa.gov domain.
    """

    base_url = "https://gpm1.gesdisc.eosdis.nasa.gov"
    file_pattern = re.compile('"[^"]*.(?:HDF5|h5|nc|nc4)"')

    def __init__(self, product):
        """
        Create new GesDisc provider.

        Args:
            product: The product to download.
        """
        self.product_name = str(product)
        self.level = self.product_name.split("_")[1][:2]
        super().__init__(product)

    @classmethod
    def get_available_products(cls):
        """
        Return the names of products available from this data provider.

        Return:
            A list of strings containing the names of the products that can
            be downloaded from this data provider.
        """
        return GPM_PRODUCTS.keys()

    @classmethod
    def download_url(cls, url, destination):
        auth = accounts.get_identity("GES DISC")
        response = requests.get(url, auth=auth)
        url = response.url
        response = requests.get(url, auth=auth)

        with open(destination, "wb") as f:
            for chunk in response:
                f.write(chunk)

    @property
    def _request_string(self):
        """The URL containing the data files for the given product."""
        base_url = "https://gpm1.gesdisc.eosdis.nasa.gov/data/"
        return base_url + GPM_PRODUCTS[str(self.product)] + "/{year}/{day}/{filename}"

    def get_files_by_year(self, year):
        """
        Return list of available files for a given day of a year.

        Args:
            year(``int``): The year for which to look up the files.

        Return:
            A list of strings containing the filenames that are available
            for the given year.
        """
        request_string = self._request_string.format(year=year, day="", filename="")
        auth = accounts.get_identity("GES DISC")
        response = requests.get(request_string, auth=auth)
        files = list(set(GesdiscProvider.file_pattern.findall(response.text)))
        return [f[1:-1] for f in files]

    def get_files_by_month(self, year, month):
        """
        Return list of available files for a given day of a year.

        Args:
            year(``int``): The year for which to look up the files.

        Return:
            A list of strings containing the filenames that are available
            for the given year.
        """
        request_string = self._request_string.format(
            year=year,
            day=f"{month:02}",
            filename=""
        )
        auth = accounts.get_identity("GES DISC")
        response = requests.get(request_string, auth=auth)
        files = list(set(GesdiscProvider.file_pattern.findall(response.text)))
        return [f[1:-1] for f in files]

    def get_files_by_day(self, year, day):
        """
        Return list of available files for a given day of a year.

        Args:
            year(``int``): The year for which to look up the files.
            day(``int``): The Julian day for which to look up the files.

        Return:
            A list of strings containing the filename that are available
            for the given day.
        """
        month = (datetime.datetime(year=year, month=1, day=1) +
                 datetime.timedelta(days=day)).month
        day = str(day)
        day = "0" * (3 - len(day)) + day
        request_string = self._request_string.format(year=year, day=day, filename="")
        auth = accounts.get_identity("GES DISC")
        print(request_string)
        response = requests.get(request_string, auth=auth)
        files = list(set(GesdiscProvider.file_pattern.findall(response.text)))
        print(files)
        if len(files) == 0:
            month = f"{month:02}"
            request_string = self._request_string.format(year=year, day=month, filename="")
            print(request_string)
            response = requests.get(request_string, auth=auth)
            files = list(set(GesdiscProvider.file_pattern.findall(response.text)))
            print(files)
        return [f[1:-1] for f in files]

    def _download_with_redirect(self, url, destination):
        """
        Handles download from GES DISC server with redirect.

        Arguments:
            url(``str``): The URL of the file to retrieve.
            destination(``str``): Destination to store the data.
        """
        auth = accounts.get_identity("GES DISC")
        response = requests.get(url, auth=auth)
        url = response.url
        response = requests.get(url, auth=auth)

        with open(destination, "wb") as f:
            for chunk in response:
                f.write(chunk)

    def download_file(self, filename, destination):
        """
        Download file from data provider.

        Args:
            filename(``str``): The name of the file to download.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """
        t = self.product.filename_to_date(filename)
        year = t.year
        day = t.strftime("%j")
        day = "0" * (3 - len(day)) + day
        if self.product.variant in ["MO"]:
            day = ""
        if self.product.variant.startswith("DAY"):
            day = f"{t.month:02}"
        url = self._request_string.format(year=year, day=day, filename=filename)
        self._download_with_redirect(url, destination)



"""
pansat.download.providers.discrete_provider
===========================================

This module providers the ``DiscreteProvider`` class, which is a base class for
providers where available cannot be determined a priori but need to be looked
up on a per-day basis.
"""
from abc import abstractmethod
from datetime import timedelta
from pathlib import Path
from pansat.download.providers.data_provider import DataProvider
import numpy as np


class DiscreteProvider(DataProvider):
    """
    The DiscreteProvider class acts as a template class for discrete data
    providers, which lookup files on a per-day basis. It only requires a
    ``get_files_by_day`` and the ``download_file`` function to be implemented
    and the extends the functionality to match the general DataProvider
    interface.
    """

    def __init__(self, product):
        super().__init__()
        self.product = product

    @abstractmethod
    def get_files_by_day(self, year, day):
        """
        Return list of available files for a given day of a year.

        Args:
            year(``int``): The year for which to look up the files.
            day(``int``): The Julian day for which to look up the files.

        Return:
            A list of strings containing the filename that are available
            for the given day.
        """

    @abstractmethod
    def download_file(self, filename, destination):
        """
        Download file from data provider.

        Args:
            filename(``str``): The name of the file to download.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """

    def download(self, start_time, end_time, destination=None):
        """
        This method downloads data for a given time range from respective the
        data provider.

        Args:
            start(``datetime.datetime``): date and time for start
            end(``datetime.datetime``): date and time for end
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """
        if not destination:
            destination = self.product.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)

        files = self.get_files_in_range(start_time, end_time)
        if len(files) == 0:
            files = self.get_files_in_range(start_time, end_time, True)

        downloaded = []
        for f in files:
            path = destination / f
            self.download_file(f, path)
            downloaded.append(path)
        return downloaded

    def get_files_in_range(self, start_time, end_time, start_inclusive=True):
        """
        Get all files within time range.

        Retrieves a list of product files that include the specified
        time range.

        Args:

            start_time(``datetime.datetime``): Start time of the time range
            end_time(``datetime.datetime``): End time of the time range
            start_inclusive(``bool``): Whether or not the list should start with
                 the first file containing ``start_time`` (True) or the first file
                 found with start time later than ``start_time`` (False).

        Returns:

            List of filename that include the specified time range.

        """
        delta = timedelta(days=1)
        time = start_time
        files = []

        year = time.year
        day = int(time.strftime("%j"))
        files_of_day = self.get_files_by_day(year, day)
        files_of_day = sorted(files_of_day, key=self.product.filename_to_date)
        time_deltas_start = np.array(
            [
                (self.product.filename_to_date(f) - start_time).total_seconds()
                for f in files_of_day
            ]
        )
        if time_deltas_start.min(initial=-1) > 0 and start_inclusive:
            previous_day = time - timedelta(days=1)
            year = previous_day.year
            day = int(previous_day.strftime("%j"))
            files_of_day = self.get_files_by_day(year, day)
            files_of_day = sorted(files_of_day, key=self.product.filename_to_date)
            files.append(files_of_day[-1])

        #
        # Go over days within range an add all included files.
        #

        while (time - end_time).total_seconds() < 24 * 60 * 60:
            if time != start_time:
                year = time.year
                day = int(time.strftime("%j"))
                files_of_day = self.get_files_by_day(year, day)
                files_of_day = sorted(files_of_day, key=self.product.filename_to_date)

            time_deltas_start = np.array(
                [
                    (self.product.filename_to_date(f) - start_time).total_seconds()
                    for f in files_of_day
                ]
            )
            indices = np.where(time_deltas_start > 0.0)[0]
            if len(indices) > 0:
                start_index = indices[0]
            else:
                start_index = -1

            time_deltas_end = np.array(
                [
                    (self.product.filename_to_date(f) - end_time).total_seconds()
                    for f in files_of_day
                ]
            )
            indices = np.where(time_deltas_end > 0.0)[0]
            if len(indices) > 0:
                end_index = indices[0]
            else:
                end_index = None

            if start_index > 0 and start_inclusive:
                start_index -= 1

            files += files_of_day[start_index:end_index]
            time += delta

        return files

    def get_file_by_date(self, time):
        """
        Get file with start time closest to a given date.

        Args:
            time(``datetime.datetime``): The date to download a file for.

        Return:
            The filename of the file with the closest start time
            before the given time.
        """
        # Check last file from previous day
        delta = timedelta(days=1)
        previous_day = time - delta
        year = previous_day.year
        day = int((previous_day.strftime("%j")))
        files = self.get_files_by_day(year, day)[-1:]

        year = time.year
        day = int(time.strftime("%j"))
        files += self.get_files_by_day(year, day)

        def time_difference(filename):
            delta = self.product.filename_to_date(filename) - time
            return delta.total_seconds()

        files_sorted = sorted(files, key=time_difference)

        def negative_only(filename):
            delta = self.product.filename_to_date(filename) - time
            return delta.total_seconds() <= 0

        files_negative = filter(negative_only, files_sorted)

        if files_negative:
            return list(files_negative)[-1]
        return files_sorted[0]



"""
====================================
pansat.download.providers.iowa_state
====================================

This module provides a data provider class to download MRMS products from an
archive located at https://mtarchive.geol.iastate.edu/, which is hosted by the
Department of Geological and atmospheric sciences at Iowa State University.

This provider doesn't require any user authentication to download data.
"""
from datetime import datetime, timedelta
import shutil

import requests

from pansat.download.providers.discrete_provider import DiscreteProvider

PRODUCTS = {
    "MRMS_PrecipRate": ["mrms", "ncep", "PrecipRate"],
    "MRMS_RadarQualityIndex": ["mrms", "ncep", "RadarQualityIndex"],
    "MRMS_PrecipFlag": ["mrms", "ncep", "PrecipFlag"],
}


class IowaStateProvider(DiscreteProvider):
    """
    Base class for data products available from htps://mtarchive.geol.iastate.edu/
    """

    base_url = "https://mtarchive.geol.iastate.edu/"

    def __init__(self, product):
        """
        Create a new product instance.

        Args:

            product(``Product``): Product class object available from the archive.

        """
        if str(product) not in PRODUCTS:
            available_products = list(PRODUCTS.keys())
            raise ValueError(
                f"The product {product} is  not a available from the Iowa State"
                f"archive provider. Currently available products are: "
                f"{available_products}."
            )
        super().__init__(product)
        self.url_suffix = "/".join(PRODUCTS[str(product)])

    @classmethod
    def get_available_products(cls):
        return PRODUCTS.keys()

    def get_url(self, date):
        url = f"{self.base_url}/{date.year}/{date.month:02}/{date.day:02}/"
        url += self.url_suffix
        return url

    def get_files_by_day(self, year, day):
        """
        Return all files from given year and julian day.

        Args:
            year(``int``): The year from which to retrieve the filenames.
            day(``int``): Day of the year of the data from which to retrieve the
                the filenames.

        Return:
            List of the filenames of this product on the given day.
        """
        date = datetime(year=year, month=1, day=1) + timedelta(days=day - 1)
        url = self.get_url(date)
        with requests.get(url) as r:
            return list(set(self.product.filename_regexp.findall(r.text)))

    def download_file(self, filename, destination):
        """
        Download file from data provider.

        Args:
            filename(``str``): The name of the file to download.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """
        date = self.product.filename_to_date(filename)
        url = self.get_url(date) + "/" + filename

        with requests.get(url, stream=True) as r:
            with open(destination, "wb") as f:
                shutil.copyfileobj(r.raw, f)



"""
pansat.products
===============

The ``products`` module provides functionality for handling supported data products.
"""

from pansat.download.providers.copernicus import COPERNICUS_PRODUCTS
from pansat.download.providers.noaa import NOAA_PRODUCTS
from pansat.download.providers.icare import ICARE_PRODUCTS
from pansat.download.providers.ges_disc import GPM_PRODUCTS


ALL_PRODUCTS = [
    *COPERNICUS_PRODUCTS,
    *NOAA_PRODUCTS,
    *list(ICARE_PRODUCTS.keys()),
    *list(GPM_PRODUCTS.keys()),
]



"""
pansat.download.providers.data_provider
=======================================

A data provider is an entity from which a specific data product can be downloaded.
A general data source, such as the Icare server or the Copernicus Climate Data
Store, are represented as classes of data providers. This module provides the
``DataProvider`` abstract base class (ABC) that defines the general interface for
data provider classes.
"""
from abc import ABCMeta, abstractmethod, abstractclassmethod


class DataProvider(metaclass=ABCMeta):
    """
    Abstract base class for data provider classes.
    """

    @abstractclassmethod
    def get_available_products(cls):
        """
        Return the names of products available from this data provider.

        This method is used by each data product to determine whether it can
        use this provider to download data. The names provided here must therefore
        match the string representation returned by the product's ``__str__`` method.

        Return:
            A list of strings containing the names of the products that can
            be downloaded from this data provider.
        """

    @abstractmethod
    def download(self, start, end, destination=None):
        """
        This method downloads data for a given time range from respective the
        data provider.

        Args:
            start(``datetime.datetime``): date and time for start
            end(``datetime.datetime``): date and time for end
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """



"""
======
pansat
======

Accessing satellite and reanalysis data made easy.
"""
import os
import logging

# Set logging level.
_LOGGING_LEVEL = os.environ.get("PANSAT_LOG_LEVEL", "WARNING")
_LOG_FORMAT = "{name} ({levelname:10}) :: {message}"
logging.basicConfig(level=_LOGGING_LEVEL, format=_LOG_FORMAT, style="{")



"""
pansat.download.providers.uci
=============================

Provider to download PERSIANN files from servers of the University
of California, Irvine.
"""
from datetime import datetime, timedelta
from urllib.request import urlopen
from pathlib import Path

from bs4 import BeautifulSoup
import requests

from pansat.download.providers.discrete_provider import DiscreteProvider

BASE_URL = "http://persiann.eng.uci.edu/CHRSdata/"
_CACHE = {}


def get_links(url):
    """
    Extract all link destination from a given URL.

    Args:
         url: URL of the website from which to retrieve the link targets.

    Return:
        List of strings containing the targets of all links find under
        the given URL.
    """
    if url not in _CACHE:
        _CACHE[url] = urlopen(url)
    _CACHE[url] = urlopen(url)
    html = _CACHE[url]
    soup = BeautifulSoup(html)
    files = []
    for link in soup.findAll('a'):
        files.append(link.get('href'))
    return files


class UciProvider(DiscreteProvider):
    """
    Data provider providing access to the PDIRNow, PERSIANN-CCS and
    PERSIANN-CDF products from the Center for Hydrometeorology and
    Remote Sensing.
    """
    @classmethod
    def get_available_products(cls):
        return [
            "PDIRNow",
            "PERSIANN-CCS",
            "PERSIANN-CDR"
        ]

    def __init__(self, product):
        """
        Create provider for a given product.
        """
        self.product = product

    def get_files_by_day(self, year, day):
        """
        Retrieve available files for a given day.

        Args:
            year: Integer specifying the year for which to retrieve the
                available files.
            day: Julian day for which to retrieve available files.

        Return:
            List of strings containing the filename of the files of the
            providers product that are available on the requested day.
        """
        url = BASE_URL + self.product.get_path(year)
        links = get_links(url)

        date = datetime(year=year, month=1, day=1) + timedelta(days=day - 1)
        files = []
        for l in links:
            try:
                f_date = self.product.filename_to_date(l)
                if ((f_date.year == date.year) and
                    (f_date.month == date.month) and
                    (f_date.day == date.day)):
                    files.append(l)
            except ValueError:
                pass
        return files

    def download_file(self, filename, destination):
        """
        Download a given file.

        Args:
            filename: Name of the file to download.
            destination: Filename to which to store the downloaded
                file.
        """
        date = self.product.filename_to_date(filename)
        year = date.year
        url = BASE_URL + self.product.get_path(year) + "/" + filename
        response = requests.get(url)
        with open(destination, "wb") as output:
            for chunk in response:
                output.write(chunk)



"""
pansat.download.providers.icare
===============================

This module providers the ``IcareProvider`` class, which implementes a data provider
class for downloading data from the
`Icare datacenter <https://www.icare.univ-lille.fr/>`_.
"""
from datetime import datetime
from ftplib import FTP
from pansat.download.providers.discrete_provider import DiscreteProvider
from pansat.download.accounts import get_identity

ICARE_PRODUCTS = {
    "CloudSat_1B-CPR": ["SPACEBORNE", "CLOUDSAT", "1B-CPR"],
    "CloudSat_2B-CLDCLASS": ["SPACEBORNE", "CLOUDSAT", "2B-CLDCLASS"],
    "CloudSat_2B-CLDCLASS-LIDAR": ["SPACEBORNE", "CLOUDSAT", "2B-CLDCLASS-LIDAR"],
    "CloudSat_2B-CWC-RO": ["SPACEBORNE", "CLOUDSAT", "2B-CWC-RO"],
    "CloudSat_2B-CWC-RVOD": ["SPACEBORNE", "CLOUDSAT", "2B-CWC-RVOD"],
    "CloudSat_2B-FLXHR": ["SPACEBORNE", "CLOUDSAT", "2B-FLXHR"],
    "CloudSat_2B-FLXHR-LIDAR": ["SPACEBORNE", "CLOUDSAT", "2B-FLXHR-LIDAR"],
    "CloudSat_2B-GEOPROF": ["SPACEBORNE", "CLOUDSAT", "2B-GEOPROF"],
    "CloudSat_2B-GEOPROF-LIDAR": ["SPACEBORNE", "CLOUDSAT", "2B-GEOPROF-LIDAR"],
    "CloudSat_2B-TAU": ["SPACEBORNE", "CLOUDSAT", "2B-TAU"],
    "CloudSat_2C-ICE": ["SPACEBORNE", "CLOUDSAT", "2B-ICE"],
    "CloudSat_2C-PRECIP-COLUMN": ["SPACEBORNE", "CLOUDSAT", "2B-PRECIP-COLUMN"],
    "CloudSat_2C-RAIN-PROFILE": ["SPACEBORNE", "CLOUDSAT", "2B-PRECIP-COLUMN"],
    "CloudSat_2C-SNOW-PROFILE": ["SPACEBORNE", "CLOUDSAT", "2B-GEOPROF-LIDAR"],
    "Calipso_333mCLay": ["SPACEBORNE", "CALIOP", "333mCLay"],
    "Calipso_01kmCLay": ["SPACEBORNE", "CALIOP", "01kmCLay"],
    "Calipso_05kmAPro": ["SPACEBORNE", "CALIOP", "05kmAPro"],
    "Calipso_CAL_LID_L1": ["SPACEBORNE", "CALIOP", "CAL_LID_L1.C3"],
    "Dardar_DARDAR-CLOUD": ["SPACEBORNE", "MULTI_SENSOR", "DARDAR-CLOUD"],
    "Dardar_DARDAR_CLOUD": ["SPACEBORNE", "MULTI_SENSOR", "DARDAR_CLOUD"],
    #"MODIS_Terra_MOD021KM": ["SPACEBORNE", "MODIS", "MOD021KM.061"],
    #"MODIS_Terra_MOD03": ["SPACEBORNE", "MODIS", "MOD03.061"],
    #"MODIS_Aqua_MYD021KM": ["SPACEBORNE", "MODIS", "MYD021KM.061"],
    #"MODIS_Aqua_MYD03": ["SPACEBORNE", "MODIS", "MYD03.061"],
    #"MODIS_Aqua_MYD35_l2": ["SPACEBORNE", "MODIS", "MYD35_L2.061"],
}


class IcareProvider(DiscreteProvider):
    """
    Base class for data products available from the ICARE ftp server.
    """

    base_url = "ftp.icare.univ-lille1.fr"

    def __init__(self, product):
        """
        Create a new product instance.

        Args:

            product(``Product``): Product class object with specific product for ICARE

        """
        if str(product) not in ICARE_PRODUCTS:
            available_products = list(ICARE_PRODUCTS.keys())
            raise ValueError(
                f"The product {product} is  not a available from the ICARE data"
                f" provider. Currently available products are: "
                f"{available_products}."
            )
        super().__init__(product)
        self.product_path = "/".join(ICARE_PRODUCTS[str(product)])
        self.cache = {}

    def _ftp_listing_to_list(self, path, item_type=int):
        """
        Retrieve directory content from ftp listing as list.

        Args:

            path(``str``): The path from which to retrieve the ftp listing.

            item_type(``type``): Type constructor to apply to the elements of the
                listing. To retrieve a list of strings use t = str.

        Return:

            A list containing the content of the ftp directory.

        """
        if not path in self.cache:
            with FTP(IcareProvider.base_url) as ftp:
                user, password = get_identity("Icare")
                ftp.login(user=user, passwd=password)
                try:
                    ftp.cwd(path)
                except:
                    raise Exception(
                        "Can't find product folder "
                        + path
                        + "on the ICARE ftp server. Are you sure this is"
                        "a ICARE multi sensor product?"
                    )
                listing = ftp.nlst()
            listing = [item_type(l) for l in listing]
            self.cache[path] = listing
        return self.cache[path]

    @classmethod
    def get_available_products(cls):
        return ICARE_PRODUCTS.keys()

    def get_files_by_day(self, year, day):
        """
        Return all files from given year and julian day.

        Args:
            year(``int``): The year from which to retrieve the filenames.
            day(``int``): Day of the year of the data from which to retrieve the
                the filenames.

        Return:
            List of the filenames of this product on the given day.
        """
        day_str = str(day)
        day_str = "0" * (3 - len(day_str)) + day_str
        date = datetime.strptime(str(year) + str(day_str), "%Y%j")
        path = "/".join([self.product_path, str(year), date.strftime("%Y_%m_%d")])
        listing = self._ftp_listing_to_list(path, str)
        files = [name for name in listing if name[-3:] == "hdf"]
        return files

    def download_file(self, filename, destination):
        """
        Download file from data provider.

        Args:
            filename(``str``): The name of the file to download.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """
        date = self.product.filename_to_date(filename)
        path = "/".join([self.product_path, str(date.year), date.strftime("%Y_%m_%d")])

        user, password = get_identity("Icare")
        with FTP(self.base_url) as ftp:
            ftp.login(user=user, passwd=password)
            ftp.cwd(path)
            with open(destination, "wb") as file:
                ftp.retrbinary("RETR " + filename, file.write)



"""
pansat.formats.hdf5
===================

This module extends the ``File`` class from the ``h5py`` module to provide
access datasets via the ``getattr`` function, which is required by the
Product description interface to turn HDF5 files into xarray Datasets.
"""
import weakref
import numpy as np

try:
    from h5py import File
except ImportError as error:
    print("The h5py package is required to read HDF5 files. Please install it.")
    raise error


class HDF5File(File):
    """
    Wrapper class for the ``h5py.File`` class that provides access to datasets
    via the ``getattribute`` method, which is used by the product description
    interface.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, attr):
        value = File.__getitem__(self, attr)
        return value



"""
pansat.download.providers
=========================

The ``providers`` module provides classes to access online services that allow
downloading data. Objects of these data provider classes provide download access
for specific data products. For this,  an object of the data provider class
must be instantiated with a given product object. If the product is available
from the data provider class, the resulting data provider instance can be used
to download files of the data product.

Example
-------

.. code-block::

    from datetime import datetime
    from pansat.download.provider.icare import IcareProvider
    from pansat.products.satellite.cloud_sat import l1b_cpr

    provider = IcareProvider(l1b_cpr)
    t_0 = datetime(2016, 11, 21, 10)
    t_1 = datetime(2016, 11, 21, 12)
    files = provider.download(t_0, t_1)

Attributes:
    ALL_PROVIDERS: List containing all available data provider classes.
        This list should be used by products to determine a data provider to
        use to download files.
"""
from pansat.download.providers.data_provider import DataProvider
from pansat.download.providers.copernicus import CopernicusProvider
from pansat.download.providers.icare import IcareProvider
from pansat.download.providers.ges_disc import GesdiscProvider
from pansat.download.providers.noaa import NOAAProvider
from pansat.download.providers.meteo_france import GeoservicesProvider
from pansat.download.providers.laads_daac import LAADSDAACProvider
from pansat.download.providers.goes_aws import GOESAWSProvider
from pansat.download.providers.iowa_state import IowaStateProvider
from pansat.download.providers.eumetsat import EUMETSATProvider
from pansat.download.providers.uci import UciProvider

ALL_PROVIDERS = [
    CopernicusProvider,
    IcareProvider,
    GesdiscProvider,
    NOAAProvider,
    GeoservicesProvider,
    LAADSDAACProvider,
    GOESAWSProvider,
    IowaStateProvider,
    EUMETSATProvider,
    UciProvider
]



"""
pansat.download.accounts
========================

The ``accounts`` sub-module handles login data for different data portals.
The login data is stored in encrypted format in a configuration
file ``identities.json`` in the user's home directory tree.

Upon first usage the ``identities.json`` file is setup with a custom user
password. This password is used to encrypt all data passwords that are
subsequently added by the user. The Fernet method is used to en- and decrypt
the passwords. All password hashing is performed using random salt.
"""
import base64
import getpass
import json
import logging
import os
from pathlib import Path

from appdirs import user_config_dir
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

# Dictionary containing provider names and corresponding user names.
_IDENTITIES = None

# The directory containing the configuration file.
_APP_DIR = Path(user_config_dir("pansat", "pansat"))
_APP_DIR.mkdir(parents=True, exist_ok=True)

# The path to the configuration file.
_IDENTITY_FILE = os.environ.get("PANSAT_IDENTITIES_FILE")
if _IDENTITY_FILE:
    _IDENTITY_FILE = Path(_IDENTITY_FILE)
else:
    _IDENTITY_FILE = Path(_APP_DIR) / Path("identities.json")

_PANSAT_SECRET = None

LOGGER = logging.getLogger(__name__)

###############################################################################
# Handling identities
###############################################################################


class WrongPasswordError(Exception):
    """
    Exception that is thrown when an entered password in incorrect.
    """


class MissingProviderError(Exception):
    """
    Exception that is thrown when no login data can be found for a given
    data provider name.
    """


def get_password(check=False):
    """
    Check if password is provided as ``PANSAT_PASSWORD`` environment variable.
    If this is not the case query user to input password.

    Params:
        check(``bool``): Whether user should insert the password twice
            to avoid spelling errors.

    Return:
        ``str`` containing the password
    """
    try:
        password = os.environ["PANSAT_PASSWORD"]
    except KeyError:
        print("Please enter your pansat user password:")
        password = getpass.getpass()

        if check:
            print("Please repeat the password:")
            password_re = getpass.getpass()
            if password != password_re:
                LOGGER.warning("The two passwords don't match!")
                get_password(check=check)

    return password


def hash_password(password, salt):
    """
    Computes the hash of a password.

    Args:
        password(``str``): The password
        salt(``bytes``): The salt to use to compute the hash.

    Return:
        The base64 encoded hash of the password.
    """
    password = password.encode()
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=salt, iterations=100000)
    return base64.urlsafe_b64encode(kdf.derive(password))


def decrypt(secret):
    """
    Decrypt password.

    Args:
        secret(``str``): The encrypted password

    Return:
        The decrypted password.

    Raises:

        cryptography.fernet.InvalidToken: Raised if the encrypted password
            is invalid.
    """
    fernet = Fernet(_PANSAT_SECRET)
    return fernet.decrypt(secret).decode("utf8")


def encrypt(password):
    """
    Encript password.

    Args:
        password(``str``): The password to encrypt.

    Return:
        The encrypted password as ``bytes``.
    """
    fernet = Fernet(_PANSAT_SECRET)
    return fernet.encrypt(password.encode())


def authenticate():
    """
    Verify pansat password.

    Reads pansat password from environment variable ``PANSAT_PASSWORD``.
    If this is not available, queries user to input password from standard
    in. Verifies password using the hash stored in the identity list.

    Raises:

        WrongPasswordError if the provided password is incorrect.
    """
    global _PANSAT_SECRET

    if _PANSAT_SECRET:
        return

    secret_hashed, salt = get_identities()["pansat"]
    secret_hashed = secret_hashed.encode()
    salt = salt.encode()

    password = get_password()

    entered_secret = hash_password(password, salt)
    entered_secret_hashed = hash_password(entered_secret.decode(), salt)

    if secret_hashed != entered_secret_hashed:
        LOGGER.error("Wrong password")
        raise WrongPasswordError("The password you entered is incorrect.")
    LOGGER.info("Authentification successful.")

    _PANSAT_SECRET = entered_secret


def initialize_identity_file():
    """
    Initializes a new identity file.

    If no existing identity file is found in the user directory, this
    function sets up a new one by generating a secret key using random
    salt and storing a hash of the key and the salt in the identity
    file.

    The salt and the hashed key can then be used to verify the user
    during subsequent logins.
    """
    global _IDENTITIES
    global _PANSAT_SECRET

    LOGGER.info("Initializing new identity file: %s", _IDENTITY_FILE)

    password = get_password(check=True)
    salt = base64.urlsafe_b64encode(os.urandom(16))
    _PANSAT_SECRET = hash_password(password, salt)

    secret_hashed = hash_password(_PANSAT_SECRET.decode(), salt)
    _IDENTITIES = {"pansat": (secret_hashed.decode(), salt.decode())}

    with open(_IDENTITY_FILE, "w") as file:
        file.write(json.dumps(_IDENTITIES))


def parse_identity_file():
    """
    If available, parses identity config file and adds entries to known
    identities.
    """
    if _IDENTITY_FILE.exists():
        global _IDENTITIES
        with open(_IDENTITY_FILE) as file:
            _IDENTITIES = json.loads(file.read())

        LOGGER.info("Parsed identity file: %s", _IDENTITY_FILE)

    else:
        initialize_identity_file()


def get_identities():
    """
    Access identity dictionary.

    Return:
        Dictionary holding identities for all known providers.
    """
    global _IDENTITIES
    if not _IDENTITIES:
        parse_identity_file()
    return _IDENTITIES


def add_identity(provider, user_name):
    """
    Add identity to known identities.

    Args:
        provider(``str``): Name of the data provider class for which the user
            name is valid.
        user(``str``): User name for the data provider.
    """
    if not _PANSAT_SECRET:
        authenticate()
    print(
        f"Please enter password for provider '{provider}' and username"
        f" '{user_name}':"
    )
    password = getpass.getpass()
    password_encrypted = encrypt(password)
    user_name_encrypted = encrypt(user_name)

    identities = get_identities()
    identities[provider] = (user_name_encrypted.decode(), password_encrypted.decode())
    with open(_IDENTITY_FILE, "w") as file:
        file.write(json.dumps(identities))

    LOGGER.info("Added identity: %s, %s", provider, user_name)


def delete_identity(provider):
    """
    Remove identity for provider.

    Args:
        provider(``str``): Name of the data provider whose identity to
            delete.
    """
    if not _PANSAT_SECRET:
        authenticate()
    identities = get_identities()
    del identities[provider]
    with open(_IDENTITY_FILE, "w") as file:
        file.write(json.dumps(identities))

    LOGGER.info("Removed identity for provider %s", provider)


def get_identity(provider):
    """
    Retrieve identity for given provider.

    Args:
        provider(``str``): Name of provider.

    Returns:
        Tuple ``(user_name, password)`` containing the user name and password
            for the given domain.

    Raises:
        MissingProviderError: if no identity for the given domain could be found.
    """
    LOGGER.info("Retrieving identity for provider %s", provider)

    if not _PANSAT_SECRET:
        authenticate()
    identities = get_identities()
    if provider in identities:
        user_name, password = identities[provider]
        password = password.encode()
        password = decrypt(password)
        user_name = user_name.encode()
        user_name = decrypt(user_name)
        return user_name, password

    raise MissingProviderError(
        f"Could not find identity for {provider}. Use the `add_identity`"
        " function to add an identity for the requested provider."
    )



"""
pansat.catalogue.local
======================

This module defines the catalogue class to look at and list information about downloaded files in local directory.

"""


import datetime
import glob
import os
from pathlib import Path
from pansat.products import ALL_PRODUCTS
from pansat.products.reanalysis import era5, ncep
from pansat.products.satellite import cloud_sat, calipso, dardar, gpm
from pansat.products.ground_based import opera


DEFAULT_DESTINATIONS = [
    "ERA5",
    "NCEP",
    "IGRA",
    "Calipso",
    "Cloudsat",
    "Dardar",
    "GPM",
    "Opera",
]


class ProductCatalogue:
    """

    The ProductCatalogue class contains methods to extract information about
    downloaded files.

    Attributes:

        available_products: string list with all currently supoprted products

    """

    def __init__(self):
        self.available_products = ALL_PRODUCTS

    def get_files_for_product(self, product, path=None):
        """

        Get list with all files for specific product.

        Args:

            product: pansat product instance
            path(``str``): string or Path for product path, if None files are listed from default destination

        Returns:
            file_lists(``list``): list containing all names that have been downloaded for a certain product.
        """

        if path == None:
            path = product.default_destination
        else:
            path = Path(path)

        file_list = []

        p = path.glob("**/*")
        files = [x for x in p if x.is_file()]

        # check whether files match respective name pattern, otherwise they are not included in list
        for f in files:
            if product.filename_regexp.match(str(f.name)):
                file_list.append(f.name)

        return file_list

    def get_file_catalogue(self, destination=None):
        """
        Getting a dictionary with all downloaded files, sorted by product and product class.

        Args:

            destination(``str``): string or Path to folder to check file structure from.
                If destination is None, the catalogue starts checking for default destinations.

        Returns:

            catalogue(``dict``): nested dictionary with all downloaded files at default locations.


        """
        catalogue = {}

        if destination == None:
            for dd in DEFAULT_DESTINATIONS:
                path = Path(dd)

                if path.is_dir():
                    p = path.glob("**/*")
                    subdirectories = [x.name for x in p if x.is_dir()]
                    subdict = {}

                    for x in subdirectories:
                        subd = path / Path(x)
                        p = subd.glob("*")
                        files = [x.name for x in p if x.is_file()]
                        subdict[str(x)] = files

                    catalogue[str(path)] = subdict

        else:
            path = Path(destination)
            p = path.glob("**/*")
            subdirectories = [x.name for x in p if x.is_dir()]
            subdict = {}

            if len(subdirectories) > 0:
                for x in subdirectories:
                    subd = path / Path(x)
                    p = subd.glob("*")
                    files = [x.name for x in p if x.is_file()]
                    subdict[str(x)] = files

            else:
                subdict = {}

            catalogue[str(path)] = subdict

        return catalogue

    def print_nested(self, d, i):
        """
        Function to print a nested dictionary in tree-like structure.
        """
        for key, value in d.items():
            print("\t" * i + str(key))
            if isinstance(value, dict):
                self.print_nested(value, i + 1)
            else:
                print("\t" * (i + 1) + str(value))

    def print_file_catalogue(self):
        """
        Prints the nested dictionary with all files per product and product class in tree-like structure.
        """
        catalogue = self.get_file_catalogue()
        self.print_nested(catalogue, 0)



"""
pansat.products.satellite.modis
===============================

This module define the MODIS product class, which is used to represent all MODIS
products.
"""
import re
from datetime import datetime
from pathlib import Path

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


class MODISProduct(Product):
    """
    Base class for MODIS product.
    """

    def __init__(self, product_name):
        self.product_name = product_name
        self.filename_regexp = re.compile(
            rf"{self.product_name.upper()}\.A(\d{{7}})\.(\d{{4}}).\d{{3}}.\d*\.hdf"
        )

    @property
    def variables(self):
        return []

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a MODIS product.

        Returns:
            ``datetime`` object representing the timestamp of the
                filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(1) + match.group(2)
        date = datetime.strptime(date_string, "%Y%j%H%M")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for CloudSat product is
        ``MODIS/<product_name>``>
        """
        return Path("MODIS")

    def __str__(self):
        """The full product name."""
        platform = "Terra"
        if self.product_name[:2] == "MY":
            platform = "Aqua"
        return f"MODIS_{platform}_{self.product_name}"

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The MODIS file to open.
        """
        raise Exception("Currently not implemented.")


modis_terra_1km = MODISProduct("MOD021KM")
modis_terra_geo = MODISProduct("MOD03")
modis_terra_cloud_mask = MODISProduct("MOD35_l2")
modis_aqua_1km = MODISProduct("MYD021KM")
modis_aqua_geo = MODISProduct("MYD03")
modis_aqua_cloud_mask = MODISProduct("MYD35_l2")



"""
=================================
pansat.products.ground_based.mrms
=================================

This module provides product class for NOAA NSSL Multi-Radar/Multi-Sensor
system (`MRMS`_).

.. _MRMS: https://www.nssl.noaa.gov/projects/mrms/

.. note::

    MRMS files come in grib2 format and therefore requires the cfgrib package
    so that they can be read using xarray.
"""
import re
from datetime import datetime
import gzip
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp

import xarray as xr

from pansat.download import providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider, MissingDependency


class MRMSProduct(Product):
    """
    This class represents MRMS products.
    """

    def __init__(self, name, variable_name):
        """
        Create new MRMS product.

        Args:
            name: The name of the MRMS product as it appears in the file signature.
            variable_name: The name to use for the data variable when opened as
                  xarray Dataset.
        """
        self.name = f"MRMS_{name}"
        self.variable_name = variable_name
        self.filename_regexp = re.compile(name + r"_00\.00_\d{8}-\d{6}.grib2\.?g?z?")

    @property
    def default_destination(self):
        """Stores MRMS files in a folder called MRMS."""
        return Path("MRMS")

    def filename_to_date(self, filename):
        """
        Extract data corresponding to MRMS file.
        """
        name = Path(filename).name.split(".")[1]
        return datetime.strptime(name, "00_%Y%m%d-%H%M%S")

    def __str__(self):
        return self.name

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open a grib2 file containing MRMS precipitation rates using xarray.

        Args:
             filename: The path to the file.

        Return:
             An xarray dataset containing the data in the file.
        """
        try:
            import cfgrib
        except ImportError as e:
            raise MissingDependency(
                f"""
                Opening grib2 files with xarray requires the cfgrib package to be installed
                and working but the following error was encountered when trying to import
                the package:

                {e}
                """
            )

        filename = Path(filename)
        if filename.suffix == ".gz":
            try:
                temp = Path(mkdtemp())
                with open(filename, "rb") as source:
                    bs = gzip.decompress(source.read())
                    dest = temp / "temp.grib2"
                    with open(dest, "wb") as dest_file:
                        dest_file.write(bs)
                    dataset = xr.load_dataset(dest, engine="cfgrib")
            finally:
                rmtree(temp)
        else:
            dataset = xr.load_dataset(filename, engine="cfgrib")
        return dataset.rename({"unknown": self.variable_name})


mrms_precip_rate = MRMSProduct("PrecipRate", "precip_rate")
mrms_radar_quality_index = MRMSProduct("RadarQualityIndex", "radar_quality_index")
mrms_precip_flag = MRMSProduct("PrecipFlag", "precip_flag")



"""
pansat.download.providers.eumetsat
==================================

This module provides the ``EUMETSATProvider`` class, which implements a data
provider class for the `EUMETSAT data store <https://www.data.eumetsat.int/>`_.

"""
import base64
from datetime import datetime
from ftplib import FTP
from pathlib import Path
import shutil

import requests

from pansat.download.providers.data_provider import DataProvider
from pansat.download.accounts import get_identity
from pansat.exceptions import CommunicationError

_PRODUCTS = {
    "MSG_Seviri": "EO:EUM:DAT:MSG:HRSEVIRI",
    "MSG_Seviri_IO": "EO:EUM:DAT:MSG:HRSEVIRI-IODC",
}


def _retrieve_access_token(key, secret):
    b64 = base64.b64encode((key + ":" + secret).encode())
    header = "Authorization: Basic " + b64.decode()
    data = {"grant_type": "client_credentials"}
    url = "http://api.eumetsat.int/token"

    r = requests.post(
        url,
        auth=requests.auth.HTTPBasicAuth(key, secret),
        data={"grant_type": "client_credentials"},
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )

    if not r.ok:
        raise CommunicationError(
            "Retrieving the EUMETSAT access token failed with the following"
            f"error:\n {r.text}"
        )

    return r.json()


class AccessToken:
    """
    Simple helper class to manage the EUMETSAT access token.
    """

    def __init__(self, key, secret):
        token_data = _retrieve_access_token(key, secret)
        self.token = token_data["access_token"]
        self.lifetime = token_data["expires_in"]
        self.created = datetime.now()

    @property
    def valid(self):
        now = datetime.now()
        age = (now - self.created).total_seconds()
        return age < self.lifetime

    def __repr__(self):
        return f"AccessToken(created={self.created}, lifetime={self.lifetime})"

    def __str__(self):
        return str(self.token)

    def renew(self, key, secret):
        self.__init__(key, secret)


class EUMETSATProvider(DataProvider):
    """
    Base class for data products available from the EUMETSAT data store.
    """

    base_url = "http://api.eumetsat.int"

    def __init__(self, product):
        """
        Create a new product instance.

        Args:

            product(``Product``): Product class object with specific product for ICARE

        """
        if str(product) not in _PRODUCTS:
            available_products = list(_PRODUCTS.keys())
            raise ValueError(
                f"The product {product} is  not a available from the Eumetsat data"
                f" provider. Currently available products are: "
                f"{available_products}."
            )
        super().__init__()
        self.product = product

    def get_available_products():
        return _PRODUCTS.keys()

    def get_files_in_range(self, start, end):
        product_id = _PRODUCTS[str(self.product)]

        parameters = {
            "format": "json",
            "pi": product_id,
            "bbox": "-90,-90, 90, 90",
            "dtstart": start.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "dtend": end.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
        }

        url = self.base_url + "/data/search-products/os"

        with requests.get(url, params=parameters) as r:
            datasets = r.json()
        links = [
            f["properties"]["links"]["data"][0]["href"] for f in datasets["features"]
        ]
        return links

    def download(self, start, end, destination):
        destination = Path(destination)

        user, password = get_identity("EUMETSAT")
        token = AccessToken(user, password)

        links = self.get_files_in_range(start, end)
        downloads = []

        for l in links:
            params = {"access_token": str(token)}
            with requests.get(l, stream=True, params=params) as r:
                if not r.ok:
                    raise CommunicationError(
                        f"Downloading the file {l} from the EUMETSAT data "
                        f"store failed token failed with the following"
                        f"error:\n {r.text}."
                    )
                filename = l.split("/")[-1]
                dest = destination / (filename + ".zip")
                with open(dest, "wb") as f:
                    shutil.copyfileobj(r.raw, f)
                downloads.append(dest)

        return downloads



"""
pansat.products.reanalysis.era5
===============================
This module defines the ERA5 product class, which represents all
supported ERA5 products.


"""
import re
import os
from datetime import datetime
from pathlib import Path

import xarray

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


class ERA5Product(Product):
    """
    The ERA5 class defines a generic interface for ERA5 products.

    Attributes:
        levels(``str``): "surface", "pressure" or "land". <surface> contains surface data
                          and column-integrated values, pressure levels contains data throughout
                          the atmosphere column and <land> contains data from surface to soil depth
        name(``str``): The full name of the product according to Copernicus webpage
        variables(``list``): List of variable names provided by this
            product.
        domain(``list``): list of strings to select region  [lat1, lat2, lon1, lon2]
                          if None: global data will be downloaded
    """

    def __init__(self, levels, variables, domain=None):

        self.variables = variables

        if not domain:
            self.domain = "global"
        else:
            if domain[0] < -90 or domain[0] > 90 or domain[1] > 90 or domain[1] < -90:
                raise Exception("Latitude values have to be between -90 and 90.")

            if (
                domain[2] < -180
                or domain[2] > 180
                or domain[3] > 180
                or domain[3] < -180
            ):
                raise Exception("Longitude values have to be between -180 and 180.")

            self.domain = domain

        if levels == "surface":
            self.levels = "single-levels"
        elif levels == "pressure":
            self.levels = "pressure-levels"
        elif levels == "land":
            self.levels = levels
        else:
            raise Exception("Supported data products are: surface, pressure and land.")

        if self.tsteps == "monthly-means":
            self.name = self.name = "reanalysis-era5-" + self.levels + "-" + self.tsteps
        elif self.tsteps == "hourly":
            self.name = "reanalysis-era5-" + self.levels

        else:
            raise Exception("tsteps has to be monthly or hourly.")

        self.filename_regexp = re.compile(
            self.name + r"_[\d]*.*_" + self.variables[0] + r".*.nc"
        )

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a ERA5 product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        filename = os.path.basename(filename)
        filename = filename.split("_")[1]
        if "monthly" in self.name:
            pattern = "%Y%m"
        else:
            pattern = "%Y%m%d%H"
        return datetime.strptime(filename, pattern)

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for ERA5 product is
        ``ERA5/<product_name>``>
        """
        return Path("ERA5") / Path(self.name)

    def __str__(self):
        """The full product name."""
        return self.name

    def download(self, start, end, destination=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start date
                of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date of the
                of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to store
                the output data.

        Returns:
            downloaded(``list``): name list of all downloaded files for data product

        """

        provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        downloaded = provider.download(start, end, destination)

        return downloaded

    @classmethod
    def open(cls, filename):
        """Opens a given file of NCEP product class as xarray.

        Args:
            filename(``str``): name of the file to be opened

        Returns:
            datasets(``xarray.Dataset``): xarray dataset object for opened file
        """

        datasets = xarray.open_dataset(filename)

        return datasets


class ERA5Monthly(ERA5Product):
    """

    Child Class of ERA5Product for monthly ERA5 data.

    """

    def __init__(self, levels, variables, domain=None):
        self.tsteps = "monthly-means"
        super().__init__(levels, variables, domain)


class ERA5Hourly(ERA5Product):
    """

    Child Class of ERA5Product for hourly ERA5 data.

    """

    def __init__(self, levels, variables, domain=None):
        self.tsteps = "hourly"
        super().__init__(levels, variables, domain)



"""
pansat.products.satellite.gpm
=============================

This module defines the GPM product class, which is used to represent all
GPM products.
"""
import re
from datetime import datetime
from pathlib import Path

import numpy as np

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.products.product_description import ProductDescription
from pansat.exceptions import NoAvailableProvider


class GPMProduct(Product):
    """
    Base class representing GPM products.
    """

    def __init__(self, level, platform, sensor, name, version, variant, description):
        self.level = level
        self.platform = platform
        self.sensor = sensor
        self.name = name
        self.version = version
        self.variant = variant
        self._description = description

        if self.variant:
            variant = "-" + self.variant
        else:
            variant = ""
        self.filename_regexp = re.compile(
            rf"{self.level}{variant}\.{self.platform}\.{self.sensor}"
            rf"\.{self.name}([\w-]*).(\d{{8}})-"
            r"S(\d{6})-E(\d{6})\.(\w*)\.((\w*)\.)?(HDF5|h5|nc|nc4)"
        )

    @property
    def variables(self):
        return []

    @property
    def description(self):
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a GPM product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        path = Path(filename)
        print(filename, self.filename_regexp)
        match = self.filename_regexp.match(path.name)

        # Some files of course have to follow a different convetion.
        if match is None:
            date_string = "20" + path.name.split("_")[2]
        else:
            date_string = match.group(2) + match.group(3)
        date = datetime.strptime(date_string, "%Y%m%d%H%M%S")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {str(self)}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for GPM product is
        ``GPM/<product_name>``>
        """
        return Path("GPM") / Path(self.name)

    def __str__(self):
        if self.variant:
            variant = f"-{self.variant}"
        else:
            variant = ""
        s = f"GPM_{self.level}{variant}_{self.platform}_{self.sensor}"
        return s

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The GPM file to open.
        """
        from pansat.formats.hdf5 import HDF5File

        file_handle = HDF5File(filename, "r")
        return self.description.to_xarray_dataset(file_handle, globals())


def _extract_scantime(scantime_group):
    """
    Extract scan time as numpy object.

    This function is use

    Args:
         scantime_group: The HDF5 group containing the scantime data.

    Returns:
         numpy.datetime64 object representing the scantime.
    """
    years = scantime_group["Year"][:]
    months = scantime_group["Month"][:]
    days = scantime_group["DayOfMonth"][:]
    hours = scantime_group["Hour"][:]
    minutes = scantime_group["Minute"][:]
    seconds = scantime_group["Second"][:]
    milli_seconds = scantime_group["MilliSecond"][:]
    n_dates = years.size
    dates = np.zeros(n_dates, dtype="datetime64[ms]")
    for i in range(n_dates):
        year = years[i]
        month = months[i]
        day = days[i]
        hour = hours[i]
        minute = minutes[i]
        second = seconds[i]
        milli_second = milli_seconds[i]
        dates[i] = np.datetime64(
            f"{year:04}-{month:02}-{day:02}"
            f"T{hour:02}:{minute:02}:{second:02}"
            f".{milli_second:03}"
        )
    return dates


def _parse_products():
    module_path = Path(__file__).parent
    for filename in module_path.iterdir():
        if filename.match("*.ini") and filename.name != "gprof.ini":
            description = ProductDescription(filename)
            python_name = filename.name.split(".")[0]
            level = description.properties["level"]
            platform = description.properties["platform"]
            sensor = description.properties["sensor"]
            name = description.properties["name"]
            version = int(description.properties["version"])
            variant = description.properties["variant"]
            globals()[python_name] = GPMProduct(
                level, platform, sensor, name, version, variant, description
            )


_parse_products()

################################################################################
# GPROF products
################################################################################


class GPROFProduct(GPMProduct):
    """
    Specialization of GPM product for GPROF products, which all have the same
    data format.
    """

    def __init__(self, platform, sensor, version, variant=""):
        module_path = Path(__file__).parent
        description = ProductDescription(module_path / "gprof.ini")
        super().__init__("2A", platform, sensor, "GPROF", version, variant, description)


l2a_gprof_gpm_gmi = GPROFProduct("GPM", "GMI", 5)
l2a_gprof_metopb_mhs = GPROFProduct("METOPB", "MHS", 5)



"""
pansat.products.satellite.calipso
=================================

This module defines the Calipso product class, which represents all
supported Calipso products.
"""
import re
import os
from datetime import datetime
from pathlib import Path

import numpy as np

from pansat.products.product_description import ProductDescription
import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


class CalipsoProduct(Product):
    """
    The Calipso class defines a generic interface for Calipso products.

    Attributes:
        name(``str``): The name of the product
        description(``list``): List of variable names provided by this
            product.
    """

    def __init__(self, name, description):
        self.name = name
        self._description = description
        self.filename_regexp = re.compile(
            r"CAL_LID_L2_"
            + name
            + r"-[\w]*-V[\d]?-[\d]{2}.([\d]{4})-([\d]{2})-([\d]{2})T([\d]{2})"
            + r"-([\d]{2})-([\d]{2})ZN\.*"
        )

    @property
    def description(self):
        # Product description object describing the Calipso product.
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a Calipso product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        filename = os.path.basename(filename)
        filename = filename.split(".")[1][:-2]
        return datetime.strptime(filename, "%Y-%m-%dT%H-%M-%S")

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for Calipso product is
        ``Calipso/<product_name>``>
        """
        return Path("Calipso") / Path(self.name)

    def __str__(self):
        """The full product name."""
        return "Calipso_" + self.name

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The Calipso file to open.
        """
        from pansat.formats.hdf4 import HDF4File

        file_handle = HDF4File(filename)
        return self.description.to_xarray_dataset(file_handle, globals())


def _parse_products():
    module_path = Path(__file__).parent
    for filename in module_path.iterdir():
        if filename.match("*.ini"):
            description = ProductDescription(filename)
            python_name = filename.name.split(".")[0]
            product_name = description.name
            globals()[python_name] = CalipsoProduct(product_name, description)


_parse_products()



"""
pansat.products.product_description
===================================

The ``product_description`` module provides a class to parse product
description files. Product descriptions are files in the ``.ini``
format that are used to describe data products. This is required for
data products that do not use the NetCDf format and thus need additional
information in order to read them into an ``xarray`` Dataset.

Example
-------

The example below provides a minimal example for a product description file.

.. code-block::

    [1B-CPR]
    type = properties
    name = 1B-CPR

    [rays]
    type = dimension

    [latitude]
    type = coordinate
    name = Latitude
    description = Spacecraft geodetic latitude.
    unit = degree
    dimensions = ["rays"]


A product description files consists of different sections, each of which
has a mandatory ``type`` defining its meaning. The following section
types are currently supported:

properties
----------

The ``properties`` section defines the general properties of the dataset.
So far, this section is used only to set the name of the product. The
title of the section is ignored.


dimension
---------

The ``dimension`` section defines a dimension used in the dataset. If
the section contains a field ``name``, then the variable of this name
will be used as numeric values defining the grid values along this
dimension. If this is not the case, ``pansat`` will automatically
try to infer the size of the dimension and use element indices
``[0, 1, ...]`` as grid values.


variable
--------

The ``variable`` section defines a multi-dimensional data array. A ``variable``
section must have fields ``name`` and ``dimensions``, which contain the name of
the corresponding data variable in a file of the product and a list of
corresponding dimensions names, respectively. Note that the dimension names
must all be defined in the same product description file.

Optionally, the ``variable`` section may contain the following additional
entries:

- ``unit``: The unit in which the data is given.
- ``description``: A string description of the variable
- ``callback``: Name of a function that is called on the corresponding
  attribute of a product file instead of trying to access the data directly
  through the attribute. This allows customized pre-processing of variable
  data.


coordinate
----------

For ``coordinate`` sections, the same rules apply as for ``variable`` section.
The difference between a coordinate and a variable is that variables are mapped
to xarray Datasets, whereas coordinates are mapped to coordinates.

Reference
---------
"""

from configparser import ConfigParser
import json

import numpy as np
import xarray


class InconsistentDimensionsError(Exception):
    """
    This error is raised when inconsistent values have been deduced for
    the length of a Dimension object.
    """


class Dimension:
    """
    Represents a dimension of multi-dimensional dataset.


    """

    def __init__(self, name, config_dict):
        """
        Create dimension object from ``.ini`` file section.

        Args:
            name(``str``): The name of the ``.ini`` file section describing the
                the dimension.
            config_dict: The dictionary describing the section of the ``.ini``
                file.
        """
        self.name = name
        self.field_name = config_dict.get("name")
        self._size = None

    @property
    def size(self):
        """The size of the dimension."""
        return self._size

    @size.setter
    def size(self, value):
        if not self._size:
            self._size = value
        else:
            if not self._size == value:
                raise InconsistentDimensionsError(
                    "Deduced inconsistent" " dimensions for dimension" f" {self.name}."
                )

    def __repr__(self):
        return f"Dimension({self.name})"


class MissingFieldError(Exception):
    """
    This error is raised when a section of an ``.ini`` file lacks a required
    key.
    """


class Variable:
    """
    Base class for all variable-type product description entries. Variable-type
    entries in a product description represent a multi-dimensional data array,
    which has a name and associated dimensions. Optionally, it may have a unit
    and a description.
    """

    def __init__(self, name, config_dict):
        """
        Args:
            name(``str``): The section name of the ``.ini`` file. which is used
                as the variable name.
            config_dict(``dict``): The dict containing the keys from the
                ``.ini`` file.
        """
        self.name = name
        self._parse_config_dict(config_dict)

    def _parse_config_dict(self, config_dict):
        """Helper function to parse config dict from ``.ini`` file."""
        if "name" not in config_dict:
            raise MissingFieldError(
                f"Dimension  definition for '{self.name}' " "has no 'name' field."
            )
        self.field_name = config_dict["name"]
        if "dimensions" not in config_dict:
            raise MissingFieldError(
                f"Dimension  definition for '{self.name}'" " has no 'dimensions' field."
            )
        self.dimensions = json.loads(config_dict["dimensions"])
        self.unit = config_dict.get("unit", "")
        self.description = config_dict.get("description", "")
        self.callback = config_dict.get("callback", [])

    def get_attributes(self):
        """
        Get dict of xarray attributes containing unit and description.
        """
        attributes = {}
        if self.unit:
            attributes["unit"] = self.unit
        if self.description:
            attributes["description"] = self.description
        return attributes

    def get_data(self, file_handle, context):
        """
        Get data for variable from file_handle.

        Retrieves data from given file handle by retrieving the attribute
        corresponding to the variables ``field_name`` attribute. If
        a callback is set for the variable, it is used to retrieve the
        data from the attribute otherwise the data is retrieved by calling
        __getitem__ with a single colon.

        Args:
             file_handle: Object providing access to a product file.
             context: Namespace in which to lookup the callback function.
        """
        if self.callback:
            callback = context[self.callback]
            data = callback(getattr(file_handle, self.field_name))
        else:
            data = getattr(file_handle, self.field_name)[:]
        return data

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


class UnknownTypeError(Exception):
    """
    This error is raised when an ``.ini`` file contains a section
    with an unknown type.
    """


class DimensionNameError(Exception):
    """
    This error is thrown when a dimension is requested that is not part
    of the product description.
    """


class ProductDescription(ConfigParser):
    """
    This class represents a description of a data product which is parsed from
    a ``.ini`` file.

    Attributes:
        name(``str``): The name of the dataset.
        dimensions: List of the dimensions of the data product.
        coordinates: List of the coordinate-type data in the data product.
        variable: List of the variable-type data in the data product.
        attribute: List of the attribute-type data in the data product.
    """

    def __init__(self, filename):
        super().__init__()
        self.dimensions = []
        self.coordinates = []
        self.variables = []
        self.attributes = []
        self._name = ""
        self.read(filename)
        self._parse_config_file()

    def _parse_config_file(self):
        for section_name in self.sections():
            section = self[section_name]
            if "type" not in section:
                raise MissingFieldError(
                    f"Section {section_name} has no 'type'" " field."
                )
            section_type = section["type"].lower()
            if section_type == "properties":
                self._parse_properties(section_name, section)
            elif section_type == "dimension":
                self.dimensions.append(Dimension(section_name, section))
            elif section_type == "coordinate":
                self.coordinates.append(Variable(section_name, section))
            elif section_type == "variable":
                self.variables.append(Variable(section_name, section))
            elif section_type == "attribute":
                self.attributes.append(Variable(section_name, section))
            else:
                raise UnknownTypeError(
                    "Type should be one of ['dimension', "
                    "'coordinate', 'variable'] but is "
                    f"'{section_type}'."
                )

    def _parse_properties(self, section_name, section):

        if "name" not in section:
            raise MissingFieldError(
                "No field 'name' in section for dimensions" f" {section_name}"
            )
        self._name = section["name"]
        self.properties = section

    @property
    def name(self):
        # The name of the data product.
        return self._name

    def _get_data(self, file_handle, context):
        """
        Reads dimensions, variables and coordinates from file_handle and
        returns them as dictionaries.

        Args:
            file_handle: File handle that provides access to the dimensions
                 and variable in this product description.

        Returns:
            A tuple ``(variables, coordinates)`` containing a two dictionaries.
            The ``variables`` maps variable names to 2-tuples of dimensionlists
            and data arrays. The ``coordinates`` dict maps coordinate names
            to dimension names or to tuples of dimension lists and data arrays.

        """
        variables = {}
        coordinates = {}
        for variable in self.variables:
            data = variable.get_data(file_handle, context)
            if len(variable.dimensions) < len(data.shape):
                data = np.squeeze(data)
            for index, dimension in enumerate(variable.dimensions):
                coordinates[dimension] = np.arange(data.shape[index])
                attrs = variable.get_attributes()
            variables[variable.name] = (variable.dimensions, data, attrs)
        for coordinate in self.coordinates:
            data = coordinate.get_data(file_handle, context)
            if len(coordinate.dimensions) < len(data.shape):
                data = np.squeeze(data)
            attrs = coordinate.get_attributes()
            coordinates[coordinate.name] = (coordinate.dimensions, data, attrs)
        return variables, coordinates

    def to_xarray_dataset(self, file_handle, context=None):
        """
        Convert data from file handle to xarray dataset.

        Args:
             file_handle: A file object providing access to the data product
                 described by this product description object.
        Return:
             ``xarray.Dataset`` containing the data from the provided file
             handle.
        """
        if not context:
            context = {}
        variables, dimensions = self._get_data(file_handle, context)
        return xarray.Dataset(variables, dimensions)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pansat",
    version="0.1",
    author="Julia Kukulies, Simon Pfreundschuh, Franz Kanngießer, Hannah Imhof",
    description="Download, extraction, remapping and analysis of satellite and climate data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SEE-MOF/pansat",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "cryptography>=3.1",
        "numpy",
        "cdsapi",
        "requests",
        "xarray",
        "pyproj",
        "appdirs",
        "boto3",
    ],
    setup_requires=["pytest-runner"],
    tests_require=[
        "sphinx_rtd_theme",
        "pytest",
        "appdirs",
        "scipy",
        "beautifulsoup4",
        "lxml",
        "netcdf4",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["pansat=pansat.download.commandline:download"]},
    include_package_data=True,
)



"""
pansat.products.reanalysis.ncep
===============================
This module defines the NCEP reanalysis product class, which represents all
supported NCEP reanalysis products.


"""

import re
import os
from datetime import datetime
from pathlib import Path

import xarray
import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


class NCEPReanalysis(Product):
    """
    The NCEP reanalysis class defines a generic interface for NCEP products.

    Attributes:
        variable(``str``): Variable to extract
        grid(``str``): pressure, surface, spectral, surface_gauss or tropopause
        name(``str``): Full name of the product.
    """

    def __init__(self, variable, grid):
        """
        Args:
            variable(``str``): Variable to extract
            grid(``str``): pressure, surface, spectral, surface_gauss or tropopause
        """

        self.variable = variable
        if grid == "tropopause":
            self.variable = variable + ".tropp"
        self.name = "ncep.reanalysis-" + str(grid)
        self.filename_regexp = re.compile(self.variable + ".*" + r".nc")

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a NCEP product.

        Returns:
            ``datetime`` object representing the timestamp of the
                filename.
        """
        filename = os.path.basename(filename)
        filename = filename.split(".")[-2]
        pattern = "%Y"

        return datetime.strptime(filename, pattern)

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find provider for the product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for NCEP product is
        ``NCEP/<product_name>``>
        """
        return Path("NCEP") / Path(self.name)

    def __str__(self):
        """The full product name."""
        return self.name

    def download(self, start, end, destination=None):
        """
        Download data product for given time range.

        Args:
            start(``int``): start year
            end(``int``): end year
            destination(``str`` or ``pathlib.Path``): The destination where to store
                the output data.

        Returns:
            downloaded(``list``): name list of all downloaded files for data product

        """

        provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        downloaded = provider.download(start, end, destination)

        return downloaded

    @classmethod
    def open(cls, filename):
        """Opens a given file of NCEP product class as xarray.

        Args:
            filename(``str``): name of the file to be opened

        Returns:
            datasets(``xarray.Dataset``): xarray dataset object for opened file
        """

        datasets = xarray.open_dataset(filename)

        return datasets



"""
pansat.download.providers.goes_aws
==================================

This module contains a data provider for data from currently operational
 GOES series satellites (GOES 16 and 17), which is available from Amazon
 AWS cloud storage.

Reference
---------
"""
from pathlib import Path

import requests
import boto3
from botocore import UNSIGNED
from botocore.config import Config

from pansat.download.providers.discrete_provider import DiscreteProvider

GOES_AWS_PRODUCTS = [
    "GOES-16-ABI-L1b-RadC",
    "GOES-16-ABI-L1b-RadF",
    "GOES-16-ABI-L1b-RadM",
    "GOES-17-ABI-L1b-RadC",
    "GOES-17-ABI-L1b-RadF",
    "GOES-17-ABI-L1b-RadM",
]

_BUCKET_CACHE = {}


class GOESAWSProvider(DiscreteProvider):
    """
    Dataprovider class for product available from NOAA GOES16 bucket on Amazon
    AWS.
    """

    bucket_name = "noaa-goes"

    def __init__(self, product):
        """
        Create new NOAA GOES16 provider.

        Args:
            product: The product to download.
        """
        super().__init__(product)
        self.product_name = str(product)[8:]  # Strip off GOES-XX
        self.bucket_name = GOESAWSProvider.bucket_name
        self.bucket_name += str(product.series_index)
        self.client = boto3.client("s3", config=Config(signature_version=UNSIGNED))

    @classmethod
    def get_available_products(cls):
        """
        Return the names of products available from this data provider.

        Return:
            A list of strings containing the names of the products that can
            be downloaded from this data provider.
        """
        return GOES_AWS_PRODUCTS

    def _get_keys(self, prefix):

        global _BUCKET_CACHE
        cache_id = (prefix, self.product.series_index)

        bucket = self.bucket_name
        kwargs = {"Bucket": bucket, "Prefix": prefix}

        files = []
        if cache_id not in _BUCKET_CACHE:
            while True:
                response = self.client.list_objects_v2(**kwargs)
                if "Contents" in response:
                    files += [Path(obj["Key"]).name for obj in response["Contents"]]
                try:
                    kwargs["ContinuationToken"] = response["NextContinuationToken"]
                except KeyError:
                    break
            _BUCKET_CACHE[cache_id] = files
        return _BUCKET_CACHE[cache_id]

    def _get_request_url(self, year, day, hour, filename):
        url = f"https://noaa-goes{self.product.series_index}.s3.amazonaws.com/"
        url += f"{self.product_name}/{year}/{day}/{hour:02}/{filename}"
        return url

    def get_files_by_day(self, year, day):
        """
        Return list of available files for a given day of a year.

        Args:
            year(``int``): The year for which to look up the files.
            day(``int``): The Julian day for which to look up the files.

        Return:
            A list of strings containing the filename that are available
            for the given day.
        """
        prefix = f"{self.product_name}/{year}/{day:03}"
        files = [f for f in self._get_keys(prefix) if self.product.matches(f)]
        return files

    def download_file(self, filename, destination):
        """
        Download file from data provider.

        Args:
            filename(``str``): The name of the file to download.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """
        t = self.product.filename_to_date(filename)
        year = t.year
        day = t.strftime("%j")
        hour = t.hour
        request_string = self._get_request_url(year, day, hour, filename)
        r = requests.get(request_string)
        with open(destination, "wb") as f:
            for chunk in r:
                f.write(chunk)



"""
pansat.products.ground_based.opera
==================================

This module defines the ``OperaProduct`` class, which is used to represent data
products from the EUMETNET Opera ground-radar network.
"""
import datetime
from pathlib import Path
import re

import numpy as np
import pyproj

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.products.product_description import ProductDescription
from pansat.exceptions import NoAvailableProvider


class OperaProduct(Product):
    """
    Class representing Opera products.

    OPERA is the Radar Programme of EUMETNET. More information can be found on
    the `EUMETNET <https://www.eumetnet.eu/activities/observations-programme/current-activities/opera/>`_ homepage.
    """

    def __init__(self, product_name, description):
        self.product_name = product_name
        self._description = description
        self.filename_regexp = re.compile(
            rf"OPERA_{self.product_name}_(\d{{4}})_(\d{{3}})"
            rf"_(\d{{2}})_(\d{{2}})\.hdf"
        )

    @property
    def variables(self):
        return []

    @property
    def description(self):
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of an Opera product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(1) + match.group(2) + match.group(3) + match.group(4)
        date = datetime.datetime.strptime(date_string, "%Y%j%H%M")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for Opera product is
        ``Opera/<product_name>``>
        """
        return Path("Opera") / Path(self.product_name)

    def __str__(self):
        s = f"OPERA_{self.product_name}"
        return s

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The Opera file to open.
        """
        from pansat.formats.hdf5 import HDF5File

        file_handle = HDF5File(filename, "r")
        return self.description.to_xarray_dataset(file_handle, globals())


def edges_to_centers(grid):
    new_grid = 0.25 * (grid[1:, 1:] + grid[1:, :-1] + grid[:-1, 1:] + grid[:-1, :-1])
    return new_grid


_LATITUDE_GRID = None
_LONGITUDE_GRID = None


def _define_coordinate_grids():
    projection_string = (
        "+proj=laea +lat_0=55.0 +lon_0=10.0 +x_0=1950000.0 "
        "+y_0=-2100000.0 +units=m +ellps=WGS84"
    )
    proj = pyproj.Proj(projection_string)
    x = np.arange(0, 1900 * 2000.0 + 1, 2000.0)
    y = np.arange(0, -2200 * 2000.0 - 1, -2000.0)
    x, y = np.meshgrid(x, y)
    lons, lats = proj(x.ravel(), y.ravel(), inverse=True)
    lats = lats.reshape(2201, 1901)
    lons = lons.reshape(2201, 1901)

    global _LATITUDE_GRID
    global _LONGITUDE_GRID
    _LATITUDE_GRID = edges_to_centers(lats)
    _LONGITUDE_GRID = edges_to_centers(lons)
    _LATITUDE_GRID.flags.writeable = False
    _LONGITUDE_GRID.flags.writeable = False


def _get_opera_projection():
    projection_string = (
        "+proj=laea +lat_0=55.0 +lon_0=10.0 +x_0=1950000.0 "
        "+y_0=-2100000.0 +units=m +ellps=WGS84"
    )
    proj = pyproj.Proj(projection_string)
    return proj


def get_latitude_grid(*args):
    """
    Returns the latitude grid on which the Opera data is provided.
    """
    global _LATITUDE_GRID
    if _LATITUDE_GRID is None:
        _define_coordinate_grids()
    return _LATITUDE_GRID


def get_longitude_grid(*args):
    """
    Returns the longitude grid on which the Opera data is provided.
    """
    global _LONGITUDE_GRID
    if _LONGITUDE_GRID is None:
        _define_coordinate_grids()
    return _LONGITUDE_GRID


def _parse_products():
    """
    Parses available Opera products.
    """
    module_path = Path(__file__).parent
    for filename in module_path.iterdir():
        if filename.match("*.ini"):
            description = ProductDescription(filename)
            python_name = filename.name.split(".")[0]
            product_name = description.name
            globals()[python_name] = OperaProduct(product_name, description)


_parse_products()



"""
pansat.download.providers.copernicus
====================================

This module provides the ``CopernicusProvider`` class to download data from the
`Copernicus data store <https://cds.climate.copernicus.eu/cdsapp#!/home>`_.
"""

import logging
from contextlib import contextmanager
import itertools
import os
from pathlib import Path
import tempfile
from datetime import timedelta
import numpy as np
import cdsapi
from pansat.download.accounts import get_identity
from pansat.download.providers.data_provider import DataProvider


COPERNICUS_PRODUCTS = [
    "reanalysis-era5-land",
    "reanalysis-era5-land-monthly-means",
    "reanalysis-era5-pressure-levels",
    "reanalysis-era5-pressure-levels-monthly-means",
    "reanalysis-era5-single-levels",
    "reanalysis-era5-single-levels-monthly-means",
]

LOGGER = logging.getLogger(__name__)


@contextmanager
def _create_cds_api_rc():
    """
    Context manager to create a temporary file with the Copernicus CDS login
    credentials obtained from the Pansat account manager.
    """
    _, path = tempfile.mkstemp()
    url, key = get_identity("Copernicus")
    # Write key to file.
    with open(path, "w") as file:
        file.write(f"url: {url}\n")
        file.write(f"key: {key}\n")

    os.environ["CDSAPI_RC"] = path
    try:
        yield path
    finally:
        os.environ.pop("CDSAPI_RC")
        Path(path).unlink()


class CopernicusProvider(DataProvider):
    """
    Base class for reanalysis products available from Copernicus.
    """

    def __init__(self, product):
        """
        Create a new product instance.

        Args:

            product(``str``): product name, available products are land, single-level,
                pressure-level for hourly and monthly resolution
        """
        super().__init__()
        self.product = product

        if not product.name in COPERNICUS_PRODUCTS:
            available_products = COPERNICUS_PRODUCTS
            raise ValueError(
                f"{product.name} not a available from the Copernicus data"
                " provider. Currently available products are: "
                f" {available_products}."
            )

    @classmethod
    def get_available_products(cls):
        """
        The products available from this dataprovider.
        """
        return COPERNICUS_PRODUCTS

    @classmethod
    def get_timesteps_monthly(cls, start, end):
        """
        Create a time range with all dates between the start and end date.

        Args:

            start(``datetime.datetime``): datetime.datetime object for start time
            end(``datetime.datetime``): datetime.datetime object for end time

        Returns:
            dates(``list``): list with months for each year
            years(``list``): list with all years

        """
        # handling data ranges over multiple years:
        if start.year != end.year:
            # get years with complete nr. of months
            full_years_range = range(start.year + 1, end.year)
            full_years = list(
                itertools.chain.from_iterable(
                    itertools.repeat(x, 12) for x in full_years_range
                )
            )
            all_months = np.arange(1, 13).astype(str)

            # get months of incomplete years
            months_first_year = list(np.arange((start.month + 1), 13).astype(str))
            months_last_year = list(np.arange(1, (end.month + 1)).astype(str))

            # create lists for years with months
            years = (
                [str(start.year)] * len(months_first_year)
                + [str(f) for f in full_years]
                + [str(end.year)] * len(months_last_year)
            )
            dates = (
                months_first_year
                + [str(m) for m in all_months] * len(full_years_range)
                + months_last_year
            )
        else:
            # getting all month for the specified year
            dates = np.arange(start.month, end.month + 1).astype(str)
            nr_of_months = np.shape(dates)[0]
            years = [str(start.year)] * nr_of_months

        return dates, years

    @classmethod
    def get_timesteps_hourly(cls, start, end):
        """
        Create a time range with all dates between the start and end date.

        Args:

            start(``datetime.datetime``): datetime.datetime object for start time
            end(``datetime.datetime``): datetime.datetime object for end time

        Returns:
            dates(``list``): list with all hours, days, months and years between two dates
        """

        delta = end - start
        dhour = delta / 3600
        dates = []
        for hour in range(dhour.seconds + 1):
            hourdelta = start + timedelta(hours=hour)
            dates.append(hourdelta)

        return dates

    def download_monthly(self, start, end, destination):
        """Downloads monthly files for given time range and stores at specified location.
        Hourly data products are saved per hour and monthly data products are
        saved per month. Note that you have to install the CDS API key before
        download is possible: https://cds.climate.copernicus.eu/api-how-to

        Args:

            start(``datetime.datetime``): start date and time (year, month, day,
                hour), if hour is not specified for hourly dataproduct, all
                hours are downloaded for each date.
            end(``datetime.datetime``): end date and time (year, month, day, hour),
                if hour is not specified for hourly dataproduct, all hours are
                downloaded for each date.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """

        # open new client instance

        with _create_cds_api_rc():
            client = cdsapi.Client()

            # subset region, if requested
            if self.product.domain == "global":
                area = ""
                domain = ""
            else:
                # change to requested order of coordinate values
                dom = np.array(
                    [
                        self.product.domain[1],
                        self.product.domain[2],
                        self.product.domain[0],
                        self.product.domain[3],
                    ]
                ).astype(str)
                area = "/".join(dom)
                domain = "-".join(np.array(self.product.domain).astype(str))

            dates, years = self.get_timesteps_monthly(start, end)
            # container to save list of downloaded files
            files = []

            # send API request for each specific month in time range
            for idx, date in enumerate(dates):
                # define download parameters for monthly download
                month = date
                year = years[idx]
                hour = "00:00"

                # zero padding for month
                if int(month) < 10:
                    month = "0" + str(month)

                filename = (
                    self.product.name
                    + "_"
                    + year
                    + month
                    + "_"
                    + "-".join(self.product.variables)
                    + domain
                    + ".nc"
                )

                # set output path and file name
                out = Path(str(destination)) / str(filename)

                # only download if file not already already exists
                if os.path.exists(out):
                    LOGGER.info(destination, "%s already exists.")
                    files.append(out)
                else:
                    if "pressure" in self.product.name:
                        client.retrieve(
                            self.product.name,
                            {
                                "product_type": "monthly_averaged_reanalysis",
                                "format": "netcdf",
                                "area": area,
                                "variable": self.product.variables,
                                "pressure_level": [
                                    "1",
                                    "2",
                                    "3",
                                    "5",
                                    "7",
                                    "10",
                                    "20",
                                    "30",
                                    "50",
                                    "70",
                                    "100",
                                    "125",
                                    "150",
                                    "175",
                                    "200",
                                    "225",
                                    "250",
                                    "300",
                                    "350",
                                    "400",
                                    "450",
                                    "500",
                                    "550",
                                    "600",
                                    "650",
                                    "700",
                                    "750",
                                    "775",
                                    "800",
                                    "825",
                                    "850",
                                    "875",
                                    "900",
                                    "925",
                                    "950",
                                    "975",
                                    "1000",
                                ],
                                "year": year,
                                "month": month,
                                "time": hour,
                            },
                            out,
                        )

                    else:
                        client.retrieve(
                            self.product.name,
                            {
                                "product_type": "monthly_averaged_reanalysis",
                                "format": "netcdf",
                                "area": area,
                                "variable": self.product.variables,
                                "year": year,
                                "month": month,
                                "time": hour,
                            },
                            out,
                        )

                    LOGGER.info("file downloaded and saved as %s", out)
                    files.append(out)

            return files

    def download_hourly(self, start, end, destination):
        """Downloads hourly files for given time range and stores at specified location.
        Hourly data products are saved per hour and monthly data products are
        saved per month. Note that you have to install the CDS API key before
        download is possible: https://cds.climate.copernicus.eu/api-how-to

        Args:

            start(``datetime.datetime``): start date and time (year, month, day,
                hour), if hour is not specified for hourly dataproduct, all
                hours are downloaded for each date.
            end(``datetime.datetime``): end date and time (year, month, day, hour),
                if hour is not specified for hourly dataproduct, all hours are
                downloaded for each date.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.
        """

        # open new client instance

        with _create_cds_api_rc():
            client = cdsapi.Client()

            # subset region, if requested
            if self.product.domain == "global":
                area = ""
                domain = ""
            else:
                # change to requested order of coordinate values
                dom = np.array(
                    [
                        self.product.domain[1],
                        self.product.domain[2],
                        self.product.domain[0],
                        self.product.domain[3],
                    ]
                ).astype(str)
                area = "/".join(dom)
                domain = "-".join(np.array(self.product.domain).astype(str))

            dates = self.get_timesteps_hourly(start, end)

            # container to save list of downloaded files
            files = []

            # send API request for each specific hour in time range
            for idx in np.arange(len(dates)):
                # define download parameters for hourly download
                year = str(dates[idx].year)
                month = str(dates[idx].month)
                day = str(dates[idx].day)
                hour = str(dates[idx].hour)

                # get download key
                download_key = "reanalysis"
                if "land" in self.product.name:
                    download_key = ""

                # zero padding for day
                if int(day) < 10:
                    day = "0" + str(day)

                # zero padding for hour string in filename
                hourstr = str(hour)
                if int(hour) < 10:
                    hourstr = "0" + str(hour)

                # zero padding for month
                if int(month) < 10:
                    month = "0" + str(month)

                filename = (
                    self.product.name
                    + "_"
                    + year
                    + month
                    + day
                    + hourstr
                    + "_"
                    + "-".join(self.product.variables)
                    + domain
                    + ".nc"
                )

                # set output path and file name
                out = Path(str(destination)) / str(filename)

                # only download if file not already already exists
                if os.path.exists(out):
                    LOGGER.info(out, "%s already exists.")
                    files.append(out)
                else:
                    if "pressure" in self.product.name:
                        client.retrieve(
                            self.product.name,
                            {
                                "product_type": download_key,
                                "format": "netcdf",
                                "area": area,
                                "variable": self.product.variables,
                                "pressure_level": [
                                    "1",
                                    "2",
                                    "3",
                                    "5",
                                    "7",
                                    "10",
                                    "20",
                                    "30",
                                    "50",
                                    "70",
                                    "100",
                                    "125",
                                    "150",
                                    "175",
                                    "200",
                                    "225",
                                    "250",
                                    "300",
                                    "350",
                                    "400",
                                    "450",
                                    "500",
                                    "550",
                                    "600",
                                    "650",
                                    "700",
                                    "750",
                                    "775",
                                    "800",
                                    "825",
                                    "850",
                                    "875",
                                    "900",
                                    "925",
                                    "950",
                                    "975",
                                    "1000",
                                ],
                                "year": year,
                                "month": month,
                                "day": day,
                                "time": hour,
                            },
                            out,
                        )

                    else:
                        client.retrieve(
                            self.product.name,
                            {
                                "product_type": download_key,
                                "format": "netcdf",
                                "area": area,
                                "variable": self.product.variables,
                                "year": year,
                                "month": month,
                                "day": day,
                                "time": hour,
                            },
                            out,
                        )

                    LOGGER.info("file downloaded and saved as %s", out)
                    files.append(out)

            return files

    def download(self, start, end, destination):
        """Downloads files dependent on desired temporal resolution of data product.

        Args:

            start(``datetime.datetime``): start date and time (year, month, day,
                hour), if hour is not specified for hourly dataproduct, all
                hours are downloaded for each date.
            end(``datetime.datetime``): end date and time (year, month, day, hour),
                if hour is not specified for hourly dataproduct, all hours are
                downloaded for each date.
            destination(``str`` or ``pathlib.Path``): path to directory where
                the downloaded files should be stored.

        """

        if "monthly" in self.product.name:
            downloaded = self.download_monthly(start, end, destination)
        else:
            downloaded = self.download_hourly(start, end, destination)

        return downloaded



"""
pansat.download.provider.scapers.open_dap
=========================================

This module implements functions to extract available products from NASA
 OpenDAP servers.
"""
from pathlib import Path
from urllib.parse import urlparse, urlunparse, urljoin
import re

import requests
from bs4 import BeautifulSoup

URL = "https://gpm1.gesdisc.eosdis.nasa.gov/opendap/"


def retrieve_page(url):
    """
    Wrapper function to retrieve URL and raise StopIteration if
    that fails.
    """
    try:
        response = requests.get(url)
    except requests.ConnectionError:
        raise StopIteration()
    return response


def is_date(text):
    """
    Determines whether link text ends on a folder representing a year or a day.

    Args:
        Link text of a link pointing to a folder on NASA OpenDAP server.

    Return:
        True is the last components of the link name represent a year or day of
        year.
    """
    try:
        last_component = text.split("/")[-2]
        year = int(last_component)
        if year > 1900 and year < 2100:
            return True
        day = int(last_component)
        return (day > 0) and (day < 366)
    except ValueError:
        return False


def map_pages(function, parent_url, depth=10):
    """
    Recursive map over child URLs of given parent URL.

    Applies a given function over all sub-domains of a give URL whose link
    name looks like a file path.

    Args:
         function: A function that is applied to the ``text`` attribute of
              the ``requests.response`` object, which contains the web page.
              Results from this function is aggregated across all found sub-
              domains.
         parent_url: The URL from which to start the scraping.


    """
    response = retrieve_page(parent_url)
    soup = BeautifulSoup(response.text, features="lxml")

    products = function(response.text)
    if products:
        results = [(parent_url, products)]
    else:
        results = []

    #
    # Search child pages.
    #

    parent_url = urlparse(parent_url)
    for link in soup.find_all("a"):
        url = urlparse(link["href"])
        if not url.netloc:
            child_path = Path(parent_url.path) / url.path
            parent_path = Path(parent_url.path)
            is_parent = parent_path in child_path.parents
            is_html = str(child_path)[-5:].lower() == ".html"
            is_folder = link.text[-1] == "/"

            if is_folder and is_parent and is_html and depth > 0:
                child_url = urljoin(urlunparse(parent_url), url.path)
                new_depth = depth - 1
                products = map_pages(function, child_url, depth=new_depth)
                results += products
                if is_date(link.text):
                    break

    return results


GPM_PRODUCT_REGEXP = re.compile(
    r'"(\w*)(?:-(\w*))?\.(\w*)\.(\w*)\..*\.HDF5"', re.MULTILINE
)


def extract_gpm_products(text):
    """
    Extract all available GPM product from given domain.

    Args:
        url: The URL of the domain to search.

    Returns:
        A list of four-tuples containing level, platform, sensor and product
        name.
    """
    found = set(re.findall(GPM_PRODUCT_REGEXP, text))
    if len(found) > 0:
        return list(found)[0]
    return []
